# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Various utilities and numeric transformations relevant to cartography.

"""

from collections import namedtuple
import copy
import warnings

import cf_units
import numpy as np
import numpy.ma as ma

import cartopy.img_transform
import cartopy.crs as ccrs
import iris.analysis
import iris.coords
import iris.coord_systems
import iris.exceptions
from iris.util import _meshgrid
from ._grid_angles import gridcell_angles, rotate_grid_vectors

# List of contents to control Sphinx autodocs.
# Unfortunately essential to get docs for the grid_angles functions.
__all__ = [
    "area_weights",
    "cosine_latitude_weights",
    "get_xy_contiguous_bounded_grids",
    "get_xy_grids",
    "gridcell_angles",
    "project",
    "rotate_grid_vectors",
    "rotate_pole",
    "rotate_winds",
    "unrotate_pole",
    "wrap_lons",
    "DistanceDifferential",
    "PartialDifferential",
]

# This value is used as a fall-back if the cube does not define the earth
DEFAULT_SPHERICAL_EARTH_RADIUS = 6367470
# TODO: This should not be necessary, as CF is always in meters
DEFAULT_SPHERICAL_EARTH_RADIUS_UNIT = cf_units.Unit("m")
# Distance differentials for coordinate systems at specified locations
DistanceDifferential = namedtuple("DistanceDifferential", "dx1 dy1 dx2 dy2")
# Partial differentials between coordinate systems
PartialDifferential = namedtuple("PartialDifferential", "dx1 dy1")


def wrap_lons(lons, base, period):
    """
    Wrap longitude values into the range between base and base+period.

    .. testsetup::

        import numpy as np
        from iris.analysis.cartography import wrap_lons

    For example:
        >>> print(wrap_lons(np.array([185, 30, -200, 75]), -180, 360))
        [-175.   30.  160.   75.]

    """
    # It is important to use 64bit floating precision when changing a floats
    # numbers range.
    lons = lons.astype(np.float64)
    return ((lons - base + period * 2) % period) + base


def unrotate_pole(rotated_lons, rotated_lats, pole_lon, pole_lat):
    """
    Convert arrays of rotated-pole longitudes and latitudes to unrotated
    arrays of longitudes and latitudes. The values of ``pole_lon`` and
    ``pole_lat`` should describe the location of the rotated pole that
    describes the arrays of rotated-pole longitudes and latitudes.

    As the arrays of rotated-pole longitudes and latitudes must describe a
    rectilinear grid, the arrays of rotated-pole longitudes and latitudes must
    be of the same shape as each other.

    Example::

        lons, lats = unrotate_pole(rotated_lons, rotated_lats, \
          pole_lon, pole_lat)

    .. note:: Uses proj.4 to perform the conversion.

    Args:

        * rotated_lons:
            An array of rotated-pole longitude values.
        * rotated_lats:
            An array of rotated-pole latitude values.
        * pole_lon:
            The longitude of the rotated pole that describes the arrays of
            rotated-pole longitudes and latitudes.
        * pole_lat:
            The latitude of the rotated pole that describes the arrays of
            rotated-pole longitudes and latitudes.

    Returns:
        An array of unrotated longitudes and an array of unrotated latitudes.

    """
    src_proj = ccrs.RotatedGeodetic(
        pole_longitude=pole_lon, pole_latitude=pole_lat
    )
    target_proj = ccrs.Geodetic()
    res = target_proj.transform_points(
        x=rotated_lons, y=rotated_lats, src_crs=src_proj
    )
    unrotated_lon = res[..., 0]
    unrotated_lat = res[..., 1]

    return unrotated_lon, unrotated_lat


def rotate_pole(lons, lats, pole_lon, pole_lat):
    """
    Convert arrays of longitudes and latitudes to arrays of rotated-pole
    longitudes and latitudes. The values of ``pole_lon`` and ``pole_lat``
    should describe the rotated pole that the arrays of longitudes and
    latitudes are to be rotated onto.

    As the arrays of longitudes and latitudes must describe a rectilinear grid,
    the arrays of rotated-pole longitudes and latitudes must be of the same
    shape as each other.

    Example::

        rotated_lons, rotated_lats = rotate_pole(lons, lats,\
         pole_lon, pole_lat)

    .. note:: Uses proj.4 to perform the conversion.

    Args:

        * lons:
            An array of longitude values.
        * lats:
            An array of latitude values.
        * pole_lon:
            The longitude of the rotated pole that the arrays of longitudes and
            latitudes are to be rotated onto.
        * pole_lat:
            The latitude of the rotated pole that the arrays of longitudes and
            latitudes are to be rotated onto.

    Returns:
        An array of rotated-pole longitudes and an array of rotated-pole
        latitudes.

    """
    src_proj = ccrs.Geodetic()
    target_proj = ccrs.RotatedGeodetic(
        pole_longitude=pole_lon, pole_latitude=pole_lat
    )
    res = target_proj.transform_points(x=lons, y=lats, src_crs=src_proj)
    rotated_lon = res[..., 0]
    rotated_lat = res[..., 1]

    return rotated_lon, rotated_lat


def _get_lon_lat_coords(cube):
    lat_coords = [
        coord for coord in cube.coords() if "latitude" in coord.name()
    ]
    lon_coords = [
        coord for coord in cube.coords() if "longitude" in coord.name()
    ]
    if len(lat_coords) > 1 or len(lon_coords) > 1:
        raise ValueError(
            "Calling `_get_lon_lat_coords` with multiple lat or lon coords"
            " is currently disallowed"
        )
    lat_coord = lat_coords[0]
    lon_coord = lon_coords[0]
    return (lon_coord, lat_coord)


def _xy_range(cube, mode=None):
    """
    Return the x & y range of this Cube.

    Args:

        * cube - The cube for which to calculate xy extents.

    Kwargs:

        * mode - If the coordinate has bounds, set this to specify the
                 min/max calculation.
                 Set to iris.coords.POINT_MODE or iris.coords.BOUND_MODE.

    """
    # Helpful error if we have an inappropriate CoordSystem
    cs = cube.coord_system("CoordSystem")
    cs_valid_types = (
        iris.coord_systems.GeogCS,
        iris.coord_systems.RotatedGeogCS,
    )
    if (cs is not None) and not isinstance(cs, cs_valid_types):
        raise ValueError(
            "Latlon coords cannot be found with {0}.".format(type(cs))
        )

    x_coord, y_coord = cube.coord(axis="X"), cube.coord(axis="Y")
    cs = cube.coord_system("CoordSystem")

    if x_coord.has_bounds() != y_coord.has_bounds():
        raise ValueError(
            "Cannot get the range of the x and y coordinates if they do "
            "not have the same presence of bounds."
        )

    if x_coord.has_bounds():
        if mode not in [iris.coords.POINT_MODE, iris.coords.BOUND_MODE]:
            raise ValueError(
                'When the coordinate has bounds, please specify "mode".'
            )
        _mode = mode
    else:
        _mode = iris.coords.POINT_MODE

    # Get the x and y grids
    if isinstance(cs, iris.coord_systems.RotatedGeogCS):
        if _mode == iris.coords.POINT_MODE:
            x, y = get_xy_grids(cube)
        else:
            x, y = get_xy_contiguous_bounded_grids(cube)
    else:
        if _mode == iris.coords.POINT_MODE:
            x = x_coord.points
            y = y_coord.points
        else:
            x = x_coord.bounds
            y = y_coord.bounds

    # Get the x and y range
    if getattr(x_coord, "circular", False):
        x_range = (np.min(x), np.min(x) + x_coord.units.modulus)
    else:
        x_range = (np.min(x), np.max(x))

    y_range = (np.min(y), np.max(y))

    return (x_range, y_range)


def get_xy_grids(cube):
    """
    Return 2D X and Y points for a given cube.

    Args:

        * cube - The cube for which to generate 2D X and Y points.

    Example::

        x, y = get_xy_grids(cube)

    """
    x_coord, y_coord = cube.coord(axis="X"), cube.coord(axis="Y")

    x = x_coord.points
    y = y_coord.points

    if x.ndim == y.ndim == 1:
        # Convert to 2D.
        x, y = _meshgrid(x, y)
    elif x.ndim == y.ndim == 2:
        # They are already in the correct shape.
        pass
    else:
        raise ValueError("Expected 1D or 2D XY coords")

    return (x, y)


def get_xy_contiguous_bounded_grids(cube):
    """
    Return 2d arrays for x and y bounds.

    Returns array of shape (n+1, m+1).

    Example::

        xs, ys = get_xy_contiguous_bounded_grids(cube)

    """
    x_coord, y_coord = cube.coord(axis="X"), cube.coord(axis="Y")

    x = x_coord.contiguous_bounds()
    y = y_coord.contiguous_bounds()
    x, y = _meshgrid(x, y)

    return (x, y)


def _quadrant_area(radian_lat_bounds, radian_lon_bounds, radius_of_earth):
    """Calculate spherical segment areas.

    - radian_lat_bounds    -- [n,2] array of latitude bounds (radians)
    - radian_lon_bounds    -- [n,2] array of longitude bounds (radians)
    - radius_of_earth      -- radius of the earth
                              (currently assumed spherical)

    Area weights are calculated for each lat/lon cell as:

        .. math::

            r^2 (lon_1 - lon_0) ( sin(lat_1) - sin(lat_0))

    The resulting array will have a shape of
    *(radian_lat_bounds.shape[0], radian_lon_bounds.shape[0])*

    The calculations are done at 64 bit precision and the returned array
    will be of type numpy.float64.

    """
    # ensure pairs of bounds
    if (
        radian_lat_bounds.shape[-1] != 2
        or radian_lon_bounds.shape[-1] != 2
        or radian_lat_bounds.ndim != 2
        or radian_lon_bounds.ndim != 2
    ):
        raise ValueError("Bounds must be [n,2] array")

    # fill in a new array of areas
    radius_sqr = radius_of_earth ** 2
    radian_lat_64 = radian_lat_bounds.astype(np.float64)
    radian_lon_64 = radian_lon_bounds.astype(np.float64)

    ylen = np.sin(radian_lat_64[:, 1]) - np.sin(radian_lat_64[:, 0])
    xlen = radian_lon_64[:, 1] - radian_lon_64[:, 0]
    areas = radius_sqr * np.outer(ylen, xlen)

    # we use abs because backwards bounds (min > max) give negative areas.
    return np.abs(areas)


def area_weights(cube, normalize=False):
    r"""
    Returns an array of area weights, with the same dimensions as the cube.

    This is a 2D lat/lon area weights array, repeated over the non lat/lon
    dimensions.

    Args:

    * cube (:class:`iris.cube.Cube`):
        The cube to calculate area weights for.

    Kwargs:

    * normalize (False/True):
        If False, weights are grid cell areas. If True, weights are grid
        cell areas divided by the total grid area.

    The cube must have coordinates 'latitude' and 'longitude' with bounds.

    Area weights are calculated for each lat/lon cell as:

        .. math::

            r^2 (lon_1 - lon_0) (\sin(lat_1) - \sin(lat_0))

    Currently, only supports a spherical datum.
    Uses earth radius from the cube, if present and spherical.
    Defaults to iris.analysis.cartography.DEFAULT_SPHERICAL_EARTH_RADIUS.

    """
    # Get the radius of the earth
    cs = cube.coord_system("CoordSystem")
    if isinstance(cs, iris.coord_systems.GeogCS):
        if cs.inverse_flattening != 0.0:
            warnings.warn("Assuming spherical earth from ellipsoid.")
        radius_of_earth = cs.semi_major_axis
    elif isinstance(cs, iris.coord_systems.RotatedGeogCS) and (
        cs.ellipsoid is not None
    ):
        if cs.ellipsoid.inverse_flattening != 0.0:
            warnings.warn("Assuming spherical earth from ellipsoid.")
        radius_of_earth = cs.ellipsoid.semi_major_axis
    else:
        warnings.warn("Using DEFAULT_SPHERICAL_EARTH_RADIUS.")
        radius_of_earth = DEFAULT_SPHERICAL_EARTH_RADIUS

    # Get the lon and lat coords and axes
    try:
        lon, lat = _get_lon_lat_coords(cube)
    except IndexError:
        raise ValueError(
            "Cannot get latitude/longitude "
            "coordinates from cube {!r}.".format(cube.name())
        )

    if lat.ndim > 1:
        raise iris.exceptions.CoordinateMultiDimError(lat)
    if lon.ndim > 1:
        raise iris.exceptions.CoordinateMultiDimError(lon)

    lat_dim = cube.coord_dims(lat)
    lat_dim = lat_dim[0] if lat_dim else None

    lon_dim = cube.coord_dims(lon)
    lon_dim = lon_dim[0] if lon_dim else None

    if not (lat.has_bounds() and lon.has_bounds()):
        msg = (
            "Coordinates {!r} and {!r} must have bounds to determine "
            "the area weights.".format(lat.name(), lon.name())
        )
        raise ValueError(msg)

    # Convert from degrees to radians
    lat = lat.copy()
    lon = lon.copy()

    for coord in (lat, lon):
        if coord.units in (cf_units.Unit("degrees"), cf_units.Unit("radians")):
            coord.convert_units("radians")
        else:
            msg = (
                "Units of degrees or radians required, coordinate "
                "{!r} has units: {!r}".format(coord.name(), coord.units.name)
            )
            raise ValueError(msg)

    # Create 2D weights from bounds.
    # Use the geographical area as the weight for each cell
    ll_weights = _quadrant_area(lat.bounds, lon.bounds, radius_of_earth)

    # Normalize the weights if necessary.
    if normalize:
        ll_weights /= ll_weights.sum()

    # Now we create an array of weights for each cell. This process will
    # handle adding the required extra dimensions and also take care of
    # the order of dimensions.
    broadcast_dims = [x for x in (lat_dim, lon_dim) if x is not None]
    wshape = []
    for idim, dim in zip((0, 1), (lat_dim, lon_dim)):
        if dim is not None:
            wshape.append(ll_weights.shape[idim])
    ll_weights = ll_weights.reshape(wshape)
    broad_weights = iris.util.broadcast_to_shape(
        ll_weights, cube.shape, broadcast_dims
    )

    return broad_weights


def cosine_latitude_weights(cube):
    r"""
    Returns an array of latitude weights, with the same dimensions as
    the cube. The weights are the cosine of latitude.

    These are n-dimensional latitude weights repeated over the dimensions
    not covered by the latitude coordinate.

    The cube must have a coordinate with 'latitude' in the name. Out of
    range values (greater than 90 degrees or less than -90 degrees) will
    be clipped to the valid range.

    Weights are calculated for each latitude as:

        .. math::

           w_l = \cos \phi_l

    Examples:

    Compute weights suitable for averaging type operations::

        from iris.analysis.cartography import cosine_latitude_weights
        cube = iris.load_cube(iris.sample_data_path('air_temp.pp'))
        weights = cosine_latitude_weights(cube)

    Compute weights suitable for EOF analysis (or other covariance type
    analyses)::

        import numpy as np
        from iris.analysis.cartography import cosine_latitude_weights
        cube = iris.load_cube(iris.sample_data_path('air_temp.pp'))
        weights = np.sqrt(cosine_latitude_weights(cube))

    """
    # Find all latitude coordinates, we want one and only one.
    lat_coords = [
        coord for coord in cube.coords() if "latitude" in coord.name()
    ]
    if len(lat_coords) > 1:
        raise ValueError("Multiple latitude coords are currently disallowed.")
    try:
        lat = lat_coords[0]
    except IndexError:
        raise ValueError(
            "Cannot get latitude "
            "coordinate from cube {!r}.".format(cube.name())
        )

    # Get the dimension position(s) of the latitude coordinate.
    lat_dims = cube.coord_dims(lat)

    # Convert to radians.
    lat = lat.copy()
    lat.convert_units("radians")

    # Compute the weights as the cosine of latitude. In some cases,
    # particularly when working in 32-bit precision, the latitude values can
    # extend beyond the allowable range of [-pi/2, pi/2] due to numerical
    # precision. We first check for genuinely out of range values, and issue a
    # warning if these are found. Then the cosine is computed and clipped to
    # the valid range [0, 1].
    threshold = np.deg2rad(0.001)  # small value for grid resolution
    if np.any(lat.points < -np.pi / 2.0 - threshold) or np.any(
        lat.points > np.pi / 2.0 + threshold
    ):
        warnings.warn(
            "Out of range latitude values will be "
            "clipped to the valid range.",
            UserWarning,
        )
    points = lat.points
    l_weights = np.cos(points).clip(0.0, 1.0)

    # Create weights for each grid point. This operation handles adding extra
    # dimensions and also the order of the dimensions.
    broadcast_dims = [x for x in lat_dims if x is not None]
    wshape = []
    for idim, dim in enumerate(lat_dims):
        if dim is not None:
            wshape.append(l_weights.shape[idim])
    l_weights = l_weights.reshape(wshape)
    broad_weights = iris.util.broadcast_to_shape(
        l_weights, cube.shape, broadcast_dims
    )

    return broad_weights


def project(cube, target_proj, nx=None, ny=None):
    """
    Nearest neighbour regrid to a specified target projection.

    Return a new cube that is the result of projecting a cube with 1 or 2
    dimensional latitude-longitude coordinates from its coordinate system into
    a specified projection e.g. Robinson or Polar Stereographic.
    This function is intended to be used in cases where the cube's coordinates
    prevent one from directly visualising the data, e.g. when the longitude
    and latitude are two dimensional and do not make up a regular grid.

    Args:
        * cube
            An instance of :class:`iris.cube.Cube`.
        * target_proj
            An instance of the Cartopy Projection class, or an instance of
            :class:`iris.coord_systems.CoordSystem` from which a projection
            will be obtained.
    Kwargs:
        * nx
            Desired number of sample points in the x direction for a domain
            covering the globe.
        * ny
            Desired number of sample points in the y direction for a domain
            covering the globe.

    Returns:
        An instance of :class:`iris.cube.Cube` and a list describing the
        extent of the projection.

    .. note::

        This function assumes global data and will if necessary extrapolate
        beyond the geographical extent of the source cube using a nearest
        neighbour approach. nx and ny then include those points which are
        outside of the target projection.

    .. note::

        Masked arrays are handled by passing their masked status to the
        resulting nearest neighbour values.  If masked, the value in the
        resulting cube is set to 0.

    .. warning::

        This function uses a nearest neighbour approach rather than any form
        of linear/non-linear interpolation to determine the data value of each
        cell in the resulting cube. Consequently it may have an adverse effect
        on the statistics of the data e.g. the mean and standard deviation
        will not be preserved.

    .. warning::

        If the target projection is non-rectangular, e.g. Robinson, the target
        grid may include points outside the boundary of the projection. The
        latitude/longitude of such points may be unpredictable.

    """
    try:
        lon_coord, lat_coord = _get_lon_lat_coords(cube)
    except IndexError:
        raise ValueError(
            "Cannot get latitude/longitude "
            "coordinates from cube {!r}.".format(cube.name())
        )

    if lat_coord.coord_system != lon_coord.coord_system:
        raise ValueError(
            "latitude and longitude coords appear to have "
            "different coordinates systems."
        )

    if lon_coord.units != "degrees":
        lon_coord = lon_coord.copy()
        lon_coord.convert_units("degrees")
    if lat_coord.units != "degrees":
        lat_coord = lat_coord.copy()
        lat_coord.convert_units("degrees")

    # Determine source coordinate system
    if lat_coord.coord_system is None:
        # Assume WGS84 latlon if unspecified
        warnings.warn(
            "Coordinate system of latitude and longitude "
            "coordinates is not specified. Assuming WGS84 Geodetic."
        )
        orig_cs = iris.coord_systems.GeogCS(
            semi_major_axis=6378137.0, inverse_flattening=298.257223563
        )
    else:
        orig_cs = lat_coord.coord_system

    # Convert to cartopy crs
    source_cs = orig_cs.as_cartopy_crs()

    # Obtain coordinate arrays (ignoring bounds) and convert to 2d
    # if not already.
    source_x = lon_coord.points
    source_y = lat_coord.points
    if source_x.ndim != 2 or source_y.ndim != 2:
        source_x, source_y = _meshgrid(source_x, source_y)

    # Calculate target grid
    target_cs = None
    if isinstance(target_proj, iris.coord_systems.CoordSystem):
        target_cs = target_proj
        target_proj = target_proj.as_cartopy_projection()

    # Resolution of new grid
    if nx is None:
        nx = source_x.shape[1]
    if ny is None:
        ny = source_x.shape[0]

    target_x, target_y, extent = cartopy.img_transform.mesh_projection(
        target_proj, nx, ny
    )

    # Determine dimension mappings - expect either 1d or 2d
    if lat_coord.ndim != lon_coord.ndim:
        raise ValueError(
            "The latitude and longitude coordinates have "
            "different dimensionality."
        )

    latlon_ndim = lat_coord.ndim
    lon_dims = cube.coord_dims(lon_coord)
    lat_dims = cube.coord_dims(lat_coord)

    if latlon_ndim == 1:
        xdim = lon_dims[0]
        ydim = lat_dims[0]
    elif latlon_ndim == 2:
        if lon_dims != lat_dims:
            raise ValueError(
                "The 2d latitude and longitude coordinates "
                "correspond to different dimensions."
            )
        # If coords are 2d assume that grid is ordered such that x corresponds
        # to the last dimension (shortest stride).
        xdim = lon_dims[1]
        ydim = lon_dims[0]
    else:
        raise ValueError(
            "Expected the latitude and longitude coordinates "
            "to have 1 or 2 dimensions, got {} and "
            "{}.".format(lat_coord.ndim, lon_coord.ndim)
        )

    # Create array to store regridded data
    new_shape = list(cube.shape)
    new_shape[xdim] = nx
    new_shape[ydim] = ny
    new_data = ma.zeros(new_shape, cube.data.dtype)

    # Create iterators to step through cube data in lat long slices
    new_shape[xdim] = 1
    new_shape[ydim] = 1
    index_it = np.ndindex(*new_shape)
    if lat_coord.ndim == 1 and lon_coord.ndim == 1:
        slice_it = cube.slices([lat_coord, lon_coord])
    elif lat_coord.ndim == 2 and lon_coord.ndim == 2:
        slice_it = cube.slices(lat_coord)
    else:
        raise ValueError(
            "Expected the latitude and longitude coordinates "
            "to have 1 or 2 dimensions, got {} and "
            "{}.".format(lat_coord.ndim, lon_coord.ndim)
        )

    #    # Mask out points outside of extent in source_cs - disabled until
    #    # a way to specify global/limited extent is agreed upon and code
    #    # is generalised to handle -180 to +180, 0 to 360 and >360 longitudes.
    #    source_desired_xy = source_cs.transform_points(target_proj,
    #                                                   target_x.flatten(),
    #                                                   target_y.flatten())
    #    if np.any(source_x < 0.0) and np.any(source_x > 180.0):
    #        raise ValueError('Unable to handle range of longitude.')
    #    # This does not work in all cases e.g. lon > 360
    #    if np.any(source_x > 180.0):
    #        source_desired_x = (source_desired_xy[:, 0].reshape(ny, nx) +
    #                            360.0) % 360.0
    #    else:
    #        source_desired_x = source_desired_xy[:, 0].reshape(ny, nx)
    #    source_desired_y = source_desired_xy[:, 1].reshape(ny, nx)
    #    outof_extent_points = ((source_desired_x < source_x.min()) |
    #                           (source_desired_x > source_x.max()) |
    #                           (source_desired_y < source_y.min()) |
    #                           (source_desired_y > source_y.max()))
    #    # Make array a mask by default (rather than a single bool) to allow mask
    #    # to be assigned to slices.
    #    new_data.mask = np.zeros(new_shape)

    # Step through cube data, regrid onto desired projection and insert results
    # in new_data array
    for index, ll_slice in zip(index_it, slice_it):
        # Regrid source data onto target grid
        index = list(index)
        index[xdim] = slice(None, None)
        index[ydim] = slice(None, None)
        index = tuple(index)  # Numpy>=1.16 : index with tuple, *not* list.
        new_data[index] = cartopy.img_transform.regrid(
            ll_slice.data,
            source_x,
            source_y,
            source_cs,
            target_proj,
            target_x,
            target_y,
        )

    #    # Mask out points beyond extent
    #    new_data[index].mask[outof_extent_points] = True

    # Remove mask if it is unnecessary
    if not np.any(new_data.mask):
        new_data = new_data.data

    # Create new cube
    new_cube = iris.cube.Cube(new_data)

    # Add new grid coords
    x_coord = iris.coords.DimCoord(
        target_x[0, :],
        "projection_x_coordinate",
        units="m",
        coord_system=copy.copy(target_cs),
    )
    y_coord = iris.coords.DimCoord(
        target_y[:, 0],
        "projection_y_coordinate",
        units="m",
        coord_system=copy.copy(target_cs),
    )

    new_cube.add_dim_coord(x_coord, xdim)
    new_cube.add_dim_coord(y_coord, ydim)

    # Add resampled lat/lon in original coord system
    source_desired_xy = source_cs.transform_points(
        target_proj, target_x.flatten(), target_y.flatten()
    )
    new_lon_points = source_desired_xy[:, 0].reshape(ny, nx)
    new_lat_points = source_desired_xy[:, 1].reshape(ny, nx)
    new_lon_coord = iris.coords.AuxCoord(
        new_lon_points,
        standard_name="longitude",
        units="degrees",
        coord_system=orig_cs,
    )
    new_lat_coord = iris.coords.AuxCoord(
        new_lat_points,
        standard_name="latitude",
        units="degrees",
        coord_system=orig_cs,
    )
    new_cube.add_aux_coord(new_lon_coord, [ydim, xdim])
    new_cube.add_aux_coord(new_lat_coord, [ydim, xdim])

    coords_to_ignore = set()
    coords_to_ignore.update(cube.coords(contains_dimension=xdim))
    coords_to_ignore.update(cube.coords(contains_dimension=ydim))
    for coord in cube.dim_coords:
        if coord not in coords_to_ignore:
            new_cube.add_dim_coord(coord.copy(), cube.coord_dims(coord))
    for coord in cube.aux_coords:
        if coord not in coords_to_ignore:
            new_cube.add_aux_coord(coord.copy(), cube.coord_dims(coord))
    discarded_coords = coords_to_ignore.difference([lat_coord, lon_coord])
    if discarded_coords:
        warnings.warn(
            "Discarding coordinates that share dimensions with "
            "{} and {}: {}".format(
                lat_coord.name(),
                lon_coord.name(),
                [coord.name() for coord in discarded_coords],
            )
        )

    # TODO handle derived coords/aux_factories

    # Copy metadata across
    new_cube.metadata = cube.metadata

    return new_cube, extent


def _transform_xy(crs_from, x, y, crs_to):
    """
    Shorthand function to transform 2d points between coordinate
    reference systems.

    Args:

    * crs_from, crs_to (:class:`cartopy.crs.Projection`):
        The coordinate reference systems.
    * x, y (arrays):
        point locations defined in 'crs_from'.

    Returns:
        x, y :  Arrays of locations defined in 'crs_to'.

    """
    pts = crs_to.transform_points(crs_from, x, y)
    return pts[..., 0], pts[..., 1]


def _inter_crs_differentials(crs1, x, y, crs2):
    """
    Calculate coordinate partial differentials from crs1 to crs2.

    Returns dx2/dx1, dy2/dx1, dx2/dy1 and dy2/dy1, at given locations.

    Args:

    * crs1, crs2 (`cartopy.crs.Projection`):
        The coordinate systems, "from" and "to".
    * x, y (array):
        Point locations defined in 'crs1'.

    Returns:
        (dx2/dx1, dy2/dx1, dx2/dy1, dy2/dy1) at given locations. Each
        element of this tuple will be the same shape as the 'x' and 'y'
        arrays and will be the partial differentials between the two systems.

    """
    # Get locations in target crs.
    crs2_x, crs2_y = _transform_xy(crs1, x, y, crs2)

    # Define small x-deltas in the source crs.
    VECTOR_DELTAS_FACTOR = 360000.0  # Empirical factor to obtain small delta.
    delta_x = (crs1.x_limits[1] - crs1.x_limits[0]) / VECTOR_DELTAS_FACTOR
    delta_x = delta_x * np.ones(x.shape)
    eps = 1e-9
    # Reverse deltas where we would otherwise step outside the valid range.
    invalid_dx = x + delta_x > crs1.x_limits[1] - eps
    delta_x[invalid_dx] = -delta_x[invalid_dx]
    # Calculate the transformed point with x = x + dx.
    crs2_x2, crs2_y2 = _transform_xy(crs1, x + delta_x, y, crs2)
    # Form differentials wrt dx.
    dx2_dx = (crs2_x2 - crs2_x) / delta_x
    dy2_dx = (crs2_y2 - crs2_y) / delta_x

    # Define small y-deltas in the source crs.
    delta_y = (crs1.y_limits[1] - crs1.y_limits[0]) / VECTOR_DELTAS_FACTOR
    delta_y = delta_y * np.ones(y.shape)
    # Reverse deltas where we would otherwise step outside the valid range.
    invalid_dy = y + delta_y > crs1.y_limits[1] - eps
    delta_y[invalid_dy] = -delta_y[invalid_dy]
    # Calculate the transformed point with y = y + dy.
    crs2_x2, crs2_y2 = _transform_xy(crs1, x, y + delta_y, crs2)
    # Form differentials wrt dy.
    dx2_dy = (crs2_x2 - crs2_x) / delta_y
    dy2_dy = (crs2_y2 - crs2_y) / delta_y

    return dx2_dx, dy2_dx, dx2_dy, dy2_dy


def _crs_distance_differentials(crs, x, y):
    """
    Calculate d(distance) / d(x) and ... / d(y) for a coordinate
    reference system at specified locations.

    Args:

    * crs (:class:`cartopy.crs.Projection`):
        The coordinate reference system.
    * x, y (array):
        Locations at which to calculate the differentials,
        defined in 'crs' coordinate reference system.

    Returns:
        (abs(ds/dx), abs(ds/dy)).
        Numerically approximated partial differentials,
        i.e. scaling factors between changes in distance and changes in
        coordinate values.

    """
    # Make a true-latlon coordinate system for distance calculations.
    crs_latlon = ccrs.Geodetic(globe=ccrs.Globe(ellipse="sphere"))
    # Transform points to true-latlon (just to get the true latitudes).
    _, true_lat = _transform_xy(crs, x, y, crs_latlon)
    # Get coordinate differentials w.r.t. true-latlon.
    dlon_dx, dlat_dx, dlon_dy, dlat_dy = _inter_crs_differentials(
        crs, x, y, crs_latlon
    )
    # Calculate effective scalings of X and Y coordinates.
    lat_factor = np.cos(np.deg2rad(true_lat)) ** 2
    ds_dx = np.sqrt(dlat_dx * dlat_dx + dlon_dx * dlon_dx * lat_factor)
    ds_dy = np.sqrt(dlat_dy * dlat_dy + dlon_dy * dlon_dy * lat_factor)
    return ds_dx, ds_dy


def _transform_distance_vectors(u_dist, v_dist, ds, dx2, dy2):
    """
    Transform distance vectors from one coordinate reference system to
    another, preserving magnitude and physical direction.

    Args:

    * u_dist, v_dist (array):
        Components of each vector along the x and y directions of the source
        crs at each location.
    * ds (`DistanceDifferential`):
        Distance differentials for the source and the target crs at specified
        locations.
    * dx2, dy2 (`PartialDifferential`):
        Partial differentials from the source to the target crs.

    Returns:
        (ut_dist, vt_dist): Tuple of arrays containing the vector components
        along the x and y directions of the target crs at each location.

    """

    # Scale input distance vectors --> source-coordinate differentials.
    u1, v1 = u_dist / ds.dx1, v_dist / ds.dy1
    # Transform vectors into the target system.
    u2 = dx2.dx1 * u1 + dx2.dy1 * v1
    v2 = dy2.dx1 * u1 + dy2.dy1 * v1
    # Rescale output coordinate vectors --> target distance vectors.
    u2_dist, v2_dist = u2 * ds.dx2, v2 * ds.dy2

    return u2_dist, v2_dist


def _transform_distance_vectors_tolerance_mask(
    src_crs, x, y, tgt_crs, ds, dx2, dy2
):
    """
    Return a mask that can be applied to data array to mask elements
    where the magnitude of vectors are not preserved due to numerical
    errors introduced by the tranformation between coordinate systems.

    Args:
    * src_crs (`cartopy.crs.Projection`):
        The source coordinate reference systems.
    * x, y (array):
        Locations of each vector defined in 'src_crs'.
    * tgt_crs (`cartopy.crs.Projection`):
        The target coordinate reference systems.
    * ds (`DistanceDifferential`):
        Distance differentials for src_crs and tgt_crs at specified locations
    * dx2, dy2 (`PartialDifferential`):
        Partial differentials from src_crs to tgt_crs.

    Returns:
        2d boolean array that is the same shape as x and y.

    """
    if x.shape != y.shape:
        raise ValueError(
            "Arrays do not have matching shapes. "
            "x.shape is {}, y.shape is {}.".format(x.shape, y.shape)
        )
    ones = np.ones(x.shape)
    zeros = np.zeros(x.shape)
    u_one_t, v_zero_t = _transform_distance_vectors(ones, zeros, ds, dx2, dy2)
    u_zero_t, v_one_t = _transform_distance_vectors(zeros, ones, ds, dx2, dy2)
    # Squared magnitudes should be equal to one within acceptable tolerance.
    # A value of atol=2e-3 is used, which corresponds to a change in magnitude
    # of approximately 0.1%.
    sqmag_1_0 = u_one_t ** 2 + v_zero_t ** 2
    sqmag_0_1 = u_zero_t ** 2 + v_one_t ** 2
    mask = np.logical_not(
        np.logical_and(
            np.isclose(sqmag_1_0, ones, atol=2e-3),
            np.isclose(sqmag_0_1, ones, atol=2e-3),
        )
    )
    return mask


def rotate_winds(u_cube, v_cube, target_cs):
    r"""
    Transform wind vectors to a different coordinate system.

    The input cubes contain U and V components parallel to the local X and Y
    directions of the input grid at each point.

    The output cubes contain the same winds, at the same locations, but
    relative to the grid directions of a different coordinate system.
    Thus in vector terms, the magnitudes will always be the same, but the
    angles can be different.

    The outputs retain the original horizontal dimension coordinates, but
    also have two 2-dimensional auxiliary coordinates containing the X and
    Y locations in the target coordinate system.

    Args:

    * u_cube
        An instance of :class:`iris.cube.Cube` that contains the x-component
        of the vector.
    * v_cube
        An instance of :class:`iris.cube.Cube` that contains the y-component
        of the vector.
    * target_cs
        An instance of :class:`iris.coord_systems.CoordSystem` that specifies
        the new grid directions.

    Returns:
        A (u', v') tuple of :class:`iris.cube.Cube` instances that are the u
        and v components in the requested target coordinate system.
        The units are the same as the inputs.

    .. note::

        The U and V values relate to distance, with units such as 'm s-1'.
        These are not the same as coordinate vectors, which transform in a
        different manner.

    .. note::

        The names of the output cubes are those of the inputs, prefixed with
        'transformed\_' (e.g. 'transformed_x_wind').

    .. warning::

        Conversion between rotated-pole and non-rotated systems can be
        expressed analytically.  However, this function always uses a numerical
        approach. In locations where this numerical approach does not preserve
        magnitude to an accuracy of 0.1%, the corresponding elements of the
        returned cubes will be masked.

    """
    # Check u_cube and v_cube have the same shape. We iterate through
    # the u and v cube slices which relies on the shapes matching.
    if u_cube.shape != v_cube.shape:
        msg = (
            "Expected u and v cubes to have the same shape. "
            "u cube has shape {}, v cube has shape {}."
        )
        raise ValueError(msg.format(u_cube.shape, v_cube.shape))

    # Check the u_cube and v_cube have the same x and y coords.
    msg = (
        "Coordinates differ between u and v cubes. Coordinate {!r} from "
        "u cube does not equal coordinate {!r} from v cube."
    )
    if u_cube.coord(axis="x") != v_cube.coord(axis="x"):
        raise ValueError(
            msg.format(
                u_cube.coord(axis="x").name(), v_cube.coord(axis="x").name()
            )
        )
    if u_cube.coord(axis="y") != v_cube.coord(axis="y"):
        raise ValueError(
            msg.format(
                u_cube.coord(axis="y").name(), v_cube.coord(axis="y").name()
            )
        )

    # Check x and y coords have the same coordinate system.
    x_coord = u_cube.coord(axis="x")
    y_coord = u_cube.coord(axis="y")
    if x_coord.coord_system != y_coord.coord_system:
        msg = (
            "Coordinate systems of x and y coordinates differ. "
            "Coordinate {!r} has a coord system of {!r}, but coordinate "
            "{!r} has a coord system of {!r}."
        )
        raise ValueError(
            msg.format(
                x_coord.name(),
                x_coord.coord_system,
                y_coord.name(),
                y_coord.coord_system,
            )
        )

    # Convert from iris coord systems to cartopy CRSs to access
    # transform functionality. Use projection as cartopy
    # transform_vectors relies on x_limits and y_limits.
    if x_coord.coord_system is not None:
        src_crs = x_coord.coord_system.as_cartopy_projection()
    else:
        # Default to Geodetic (but actually use PlateCarree as a
        # projection is needed).
        src_crs = ccrs.PlateCarree()
    target_crs = target_cs.as_cartopy_projection()

    # Check the number of dimensions of the x and y coords is the same.
    # Subsequent logic assumes either both 1d or both 2d.
    x = x_coord.points
    y = y_coord.points
    if x.ndim != y.ndim or x.ndim > 2 or y.ndim > 2:
        msg = (
            "x and y coordinates must have the same number of dimensions "
            "and be either 1D or 2D. The number of dimensions are {} and "
            "{}, respectively.".format(x.ndim, y.ndim)
        )
        raise ValueError(msg)

    # Check the dimension mappings match between u_cube and v_cube.
    if u_cube.coord_dims(x_coord) != v_cube.coord_dims(x_coord):
        raise ValueError(
            "Dimension mapping of x coordinate differs "
            "between u and v cubes."
        )
    if u_cube.coord_dims(y_coord) != v_cube.coord_dims(y_coord):
        raise ValueError(
            "Dimension mapping of y coordinate differs "
            "between u and v cubes."
        )
    x_dims = u_cube.coord_dims(x_coord)
    y_dims = u_cube.coord_dims(y_coord)

    # Convert points to 2D, if not already, and determine dims.
    if x.ndim == y.ndim == 1:
        x, y = _meshgrid(x, y)
        dims = (y_dims[0], x_dims[0])
    else:
        dims = x_dims

    # Transpose x, y 2d arrays to match the order in cube's data
    # array so that x, y and the sliced data all line up.
    if dims[0] > dims[1]:
        x = x.transpose()
        y = y.transpose()

    # Create resulting cubes.
    ut_cube = u_cube.copy()
    vt_cube = v_cube.copy()
    ut_cube.rename("transformed_{}".format(u_cube.name()))
    vt_cube.rename("transformed_{}".format(v_cube.name()))

    # Get distance scalings for source crs.
    ds_dx1, ds_dy1 = _crs_distance_differentials(src_crs, x, y)

    # Get distance scalings for target crs.
    x2, y2 = _transform_xy(src_crs, x, y, target_crs)
    ds_dx2, ds_dy2 = _crs_distance_differentials(target_crs, x2, y2)

    ds = DistanceDifferential(ds_dx1, ds_dy1, ds_dx2, ds_dy2)

    # Calculate coordinate partial differentials from source crs to target crs.
    dx2_dx1, dy2_dx1, dx2_dy1, dy2_dy1 = _inter_crs_differentials(
        src_crs, x, y, target_crs
    )

    dx2 = PartialDifferential(dx2_dx1, dx2_dy1)
    dy2 = PartialDifferential(dy2_dx1, dy2_dy1)

    # Calculate mask based on preservation of magnitude.
    mask = _transform_distance_vectors_tolerance_mask(
        src_crs, x, y, target_crs, ds, dx2, dy2
    )
    apply_mask = mask.any()
    if apply_mask:
        # Make masked arrays to accept masking.
        ut_cube.data = ma.asanyarray(ut_cube.data)
        vt_cube.data = ma.asanyarray(vt_cube.data)

    # Project vectors with u, v components one horiz slice at a time and
    # insert into the resulting cubes.
    shape = list(u_cube.shape)
    for dim in dims:
        shape[dim] = 1
    ndindex = np.ndindex(*shape)
    for index in ndindex:
        index = list(index)
        for dim in dims:
            index[dim] = slice(None, None)
        index = tuple(index)
        u = u_cube.data[index]
        v = v_cube.data[index]
        ut, vt = _transform_distance_vectors(u, v, ds, dx2, dy2)
        if apply_mask:
            ut = ma.asanyarray(ut)
            ut[mask] = ma.masked
            vt = ma.asanyarray(vt)
            vt[mask] = ma.masked
        ut_cube.data[index] = ut
        vt_cube.data[index] = vt

    # Calculate new coords of locations in target coordinate system.
    xyz_tran = target_crs.transform_points(src_crs, x, y)
    xt = xyz_tran[..., 0].reshape(x.shape)
    yt = xyz_tran[..., 1].reshape(y.shape)

    # Transpose xt, yt 2d arrays to match the dim order
    # of the original x an y arrays - i.e. undo the earlier
    # transpose (if applied).
    if dims[0] > dims[1]:
        xt = xt.transpose()
        yt = yt.transpose()

    xt_coord = iris.coords.AuxCoord(
        xt, standard_name="projection_x_coordinate", coord_system=target_cs
    )
    yt_coord = iris.coords.AuxCoord(
        yt, standard_name="projection_y_coordinate", coord_system=target_cs
    )
    # Set units based on coord_system.
    if isinstance(
        target_cs,
        (iris.coord_systems.GeogCS, iris.coord_systems.RotatedGeogCS),
    ):
        xt_coord.units = yt_coord.units = "degrees"
    else:
        xt_coord.units = yt_coord.units = "m"

    ut_cube.add_aux_coord(xt_coord, dims)
    ut_cube.add_aux_coord(yt_coord, dims)
    vt_cube.add_aux_coord(xt_coord.copy(), dims)
    vt_cube.add_aux_coord(yt_coord.copy(), dims)

    return ut_cube, vt_cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp.load` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import iris
from iris.fileformats.pp import _convert_constraints
from iris.fileformats.pp import STASH


class Test_convert_constraints(tests.IrisTest):
    def _single_stash(self):
        constraint = iris.AttributeConstraint(STASH="m01s03i236")
        return _convert_constraints(constraint)

    def test_single_stash(self):
        pp_filter = self._single_stash()
        stcube = mock.Mock(stash=STASH.from_msi("m01s03i236"))
        self.assertTrue(pp_filter(stcube))

    def test_stash_object(self):
        constraint = iris.AttributeConstraint(
            STASH=STASH.from_msi("m01s03i236")
        )
        pp_filter = _convert_constraints(constraint)
        stcube = mock.Mock(stash=STASH.from_msi("m01s03i236"))
        self.assertTrue(pp_filter(stcube))

    def test_surface_altitude(self):
        # Ensure that surface altitude fields are not filtered.
        pp_filter = self._single_stash()
        orography_cube = mock.Mock(stash=STASH.from_msi("m01s00i033"))
        self.assertTrue(pp_filter(orography_cube))

    def test_surface_pressure(self):
        # Ensure that surface pressure fields are not filtered.
        pp_filter = self._single_stash()
        pressure_cube = mock.Mock(stash=STASH.from_msi("m01s00i001"))
        self.assertTrue(pp_filter(pressure_cube))

    def test_double_stash(self):
        stcube236 = mock.Mock(stash=STASH.from_msi("m01s03i236"))
        stcube4 = mock.Mock(stash=STASH.from_msi("m01s00i004"))
        stcube7 = mock.Mock(stash=STASH.from_msi("m01s00i007"))
        constraints = [
            iris.AttributeConstraint(STASH="m01s03i236"),
            iris.AttributeConstraint(STASH="m01s00i004"),
        ]
        pp_filter = _convert_constraints(constraints)
        self.assertTrue(pp_filter(stcube236))
        self.assertTrue(pp_filter(stcube4))
        self.assertFalse(pp_filter(stcube7))

    def test_callable_stash(self):
        stcube236 = mock.Mock(stash=STASH.from_msi("m01s03i236"))
        stcube4 = mock.Mock(stash=STASH.from_msi("m01s00i004"))
        stcube7 = mock.Mock(stash=STASH.from_msi("m01s00i007"))
        con1 = iris.AttributeConstraint(STASH=lambda s: s.endswith("004"))
        con2 = iris.AttributeConstraint(STASH=lambda s: s == "m01s00i007")
        constraints = [con1, con2]
        pp_filter = _convert_constraints(constraints)
        self.assertFalse(pp_filter(stcube236))
        self.assertTrue(pp_filter(stcube4))
        self.assertTrue(pp_filter(stcube7))

    def test_multiple_with_stash(self):
        constraints = [
            iris.Constraint("air_potential_temperature"),
            iris.AttributeConstraint(STASH="m01s00i004"),
        ]
        pp_filter = _convert_constraints(constraints)
        self.assertIsNone(pp_filter)

    def test_no_stash(self):
        constraints = [
            iris.Constraint("air_potential_temperature"),
            iris.AttributeConstraint(source="asource"),
        ]
        pp_filter = _convert_constraints(constraints)
        self.assertIsNone(pp_filter)

    def test_no_constraint(self):
        constraints = []
        pp_filter = _convert_constraints(constraints)
        self.assertIsNone(pp_filter)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._convert_pseudo_level_coords`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.coords import DimCoord
from iris.tests.unit.fileformats import TestField

from iris.fileformats.pp_load_rules import _convert_scalar_pseudo_level_coords


class Test(TestField):
    def test_valid(self):
        coords_and_dims = _convert_scalar_pseudo_level_coords(lbuser5=21)
        self.assertEqual(
            coords_and_dims,
            [(DimCoord([21], long_name="pseudo_level", units="1"), None)],
        )

    def test_missing_indicator(self):
        coords_and_dims = _convert_scalar_pseudo_level_coords(lbuser5=0)
        self.assertEqual(coords_and_dims, [])


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Common test code for `iris.plot.pcolor` and `iris.plot.pcolormesh`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.tests.stock import simple_2d
from iris.tests.unit.plot import MixinCoords


class MixinStringCoordPlot:
    # Mixin for common string-coord tests on pcolor/pcolormesh.
    # To use, make a class that inherits from this *and*
    # :class:`iris.tests.unit.plot.TestGraphicStringCoord`,
    # and defines "self.blockplot_func()", to return the `iris.plot` function.
    def test_yaxis_labels(self):
        self.blockplot_func()(self.cube, coords=("bar", "str_coord"))
        self.assertBoundsTickLabels("yaxis")

    def test_xaxis_labels(self):
        self.blockplot_func()(self.cube, coords=("str_coord", "bar"))
        self.assertBoundsTickLabels("xaxis")

    def test_xaxis_labels_with_axes(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlim(0, 3)
        self.blockplot_func()(self.cube, coords=("str_coord", "bar"), axes=ax)
        plt.close(fig)
        self.assertPointsTickLabels("xaxis", ax)

    def test_yaxis_labels_with_axes(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_ylim(0, 3)
        self.blockplot_func()(self.cube, axes=ax, coords=("bar", "str_coord"))
        plt.close(fig)
        self.assertPointsTickLabels("yaxis", ax)

    def test_geoaxes_exception(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        self.assertRaises(
            TypeError, self.blockplot_func(), self.lat_lon_cube, axes=ax
        )
        plt.close(fig)


class Mixin2dCoordsPlot(MixinCoords):
    # Mixin for common coordinate tests on pcolor/pcolormesh.
    # To use, make a class that inherits from this *and*
    # :class:`iris.tests.IrisTest`,
    # and defines "self.blockplot_func()", to return the `iris.plot` function.
    def blockplot_setup(self):
        # We have a 2d cube with dimensionality (bar: 3; foo: 4)
        self.cube = simple_2d(with_bounds=True)
        coord = self.cube.coord("foo")
        self.foo = coord.contiguous_bounds()
        self.foo_index = np.arange(coord.points.size + 1)
        coord = self.cube.coord("bar")
        self.bar = coord.contiguous_bounds()
        self.bar_index = np.arange(coord.points.size + 1)
        self.data = self.cube.data
        self.dataT = self.data.T
        self.draw_func = self.blockplot_func()
        patch_target_name = "matplotlib.pyplot." + self.draw_func.__name__
        self.mpl_patch = self.patch(patch_target_name)


class Mixin2dCoordsContigTol:
    # Mixin for contiguity tolerance argument to pcolor/pcolormesh.
    # To use, make a class that inherits from this *and*
    # :class:`iris.tests.IrisTest`,
    # and defines "self.blockplot_func()", to return the `iris.plot` function,
    # and defines "self.additional_kwargs" for expected extra call args.
    def test_contig_tol(self):
        # Patch the inner call to ensure contiguity_tolerance is passed.
        cube_argument = mock.sentinel.passed_arg
        expected_result = mock.sentinel.returned_value
        blockplot_patch = self.patch(
            "iris.plot._draw_2d_from_bounds",
            mock.Mock(return_value=expected_result),
        )
        # Make the call
        draw_func = self.blockplot_func()
        other_kwargs = self.additional_kwargs
        result = draw_func(cube_argument, contiguity_tolerance=0.0123)
        drawfunc_name = draw_func.__name__
        # Check details of the call that was made.
        self.assertEqual(
            blockplot_patch.call_args_list,
            [
                mock.call(
                    drawfunc_name,
                    cube_argument,
                    contiguity_tolerance=0.0123,
                    **other_kwargs,
                )
            ],
        )
        self.assertEqual(result, expected_result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# -*- coding: utf-8 -*-
#
# Iris documentation build configuration file, created by
# sphinx-quickstart on Tue May 25 13:26:23 2010.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# ----------------------------------------------------------------------------

import ntpath
import os
import sys


# function to write  useful output to stdout, prefixing the source.
def autolog(message):
    print("[{}] {}".format(ntpath.basename(__file__), message))


# -- Are we running on the readthedocs server, if so do some setup -----------

on_rtd = os.environ.get("READTHEDOCS") == "True"

if on_rtd:
    autolog("Build running on READTHEDOCS server")

    # list all the READTHEDOCS environment variables that may be of use
    # at some point
    autolog("Listing all environment variables on the READTHEDOCS server...")

    for item, value in os.environ.items():
        autolog("[READTHEDOCS] {} = {}".format(item, value))

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime
import warnings

# custom sphinx extensions
sys.path.append(os.path.abspath("sphinxext"))

# add some sample files from the developers guide..
sys.path.append(os.path.abspath(os.path.join("developers_guide")))

# why isnt the iris path added to it is discoverable too?  We dont need to,
# the sphinext to generate the api rst knows where the source is.  If it
# is added then the travis build will likely fail.

# -- Project information -----------------------------------------------------

project = "Iris"

# define the copyright information for latex builds. Note, for html builds,
# the copyright exists directly inside "_templates/layout.html"
upper_copy_year = datetime.datetime.now().year
copyright = "Iris Contributors"
author = "Iris Developers"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

import iris

# The short X.Y version.
if iris.__version__ == "dev":
    version = "dev"
else:
    # major.minor.patch-dev -> major.minor.patch
    version = ".".join(iris.__version__.split("-")[0].split(".")[:3])
# The full version, including alpha/beta/rc tags.
release = iris.__version__

autolog("Iris Version = {}".format(version))
autolog("Iris Release = {}".format(release))

# -- General configuration ---------------------------------------------------

# Create a variable that can be insterted in the rst "|copyright_years|".
# You can add more vairables here if needed
rst_epilog = """
.. |copyright_years| replace:: {year_range}
""".format(
    year_range="2010 - {}".format(upper_copy_year)
)

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.duration",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx_panels",
    # TODO: Spelling extension disabled until the dependencies can be included
    # "sphinxcontrib.spelling",
    "sphinx_gallery.gen_gallery",
    "matplotlib.sphinxext.mathmpl",
    "matplotlib.sphinxext.plot_directive",
    # better api documentation (custom)
    "custom_class_autodoc",
    "custom_data_autodoc",
    "generate_package_rst",
]

# -- panels extension ---------------------------------------------------------
# See https://sphinx-panels.readthedocs.io/en/latest/

# -- Napoleon extension -------------------------------------------------------
# See https://sphinxcontrib-napoleon.readthedocs.io/en/latest/sphinxcontrib.napoleon.html
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True  # includes dunders in api doc
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_use_keyword = True
napoleon_custom_sections = None

# -- spellingextension --------------------------------------------------------
# See https://sphinxcontrib-spelling.readthedocs.io/en/latest/customize.html
spelling_lang = "en_GB"
# The lines in this file must only use line feeds (no carriage returns).
spelling_word_list_filename = ["spelling_allow.txt"]
spelling_show_suggestions = False
spelling_show_whole_line = False
spelling_ignore_importable_modules = True
spelling_ignore_python_builtins = True

# -- copybutton extension -----------------------------------------------------
# See https://sphinx-copybutton.readthedocs.io/en/latest/
copybutton_prompt_text = ">>> "

# sphinx.ext.todo configuration
# See https://www.sphinx-doc.org/en/master/usage/extensions/todo.html
todo_include_todos = True

# api generation configuration
autodoc_member_order = "groupwise"
autodoc_default_flags = ["show-inheritance"]
autosummary_generate = True
autosummary_imported_members = True
autopackage_name = ["iris"]
autoclass_content = "init"
modindex_common_prefix = ["iris"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# -- intersphinx extension ----------------------------------------------------
# See https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    "cartopy": ("https://scitools.org.uk/cartopy/docs/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
}

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- plot_directive extension -------------------------------------------------
# See https://matplotlib.org/stable/api/sphinxext_plot_directive_api.html#options
plot_formats = [
    ("png", 100),
]

# -- Extlinks extension -------------------------------------------------------
# See https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html

extlinks = {
    "issue": ("https://github.com/SciTools/iris/issues/%s", "Issue #"),
    "pull": ("https://github.com/SciTools/iris/pull/%s", "PR #"),
}

# -- Doctest ("make doctest")--------------------------------------------------

doctest_global_setup = "import iris"

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_logo = "_static/iris-logo-title.png"
html_favicon = "_static/favicon.ico"
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "display_version": True,
    "style_external_links": True,
    "logo_only": "True",
}

html_context = {
    "copyright_years": "2010 - {}".format(upper_copy_year),
    # menu_links and menu_links_name are used in _templates/layout.html
    # to include some nice icons.  See http://fontawesome.io for a list of
    # icons (used in the sphinx_rtd_theme)
    "menu_links_name": "Support",
    "menu_links": [
        (
            '<i class="fa fa-github fa-fw"></i> Source Code',
            "https://github.com/SciTools/iris",
        ),
        (
            '<i class="fa fa-comments fa-fw"></i> Users Google Group',
            "https://groups.google.com/forum/#!forum/scitools-iris",
        ),
        (
            '<i class="fa fa-comments fa-fw"></i> Developers Google Group',
            "https://groups.google.com/forum/#!forum/scitools-iris-dev",
        ),
        (
            '<i class="fa fa-question fa-fw"></i> StackOverflow for "How Do I?"',
            "https://stackoverflow.com/questions/tagged/python-iris",
        ),
        (
            '<i class="fa fa-book fa-fw"></i> Legacy Documentation',
            "https://scitools.org.uk/iris/docs/v2.4.0/index.html",
        ),
    ],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_style = "theme_override.css"

# url link checker.  Some links work but report as broken, lets ignore them.
# See https://www.sphinx-doc.org/en/1.2/config.html#options-for-the-linkcheck-builder
linkcheck_ignore = [
    "http://cfconventions.org",
    "http://code.google.com/p/msysgit/downloads/list",
    "http://effbot.org",
    "https://github.com",
    "http://www.personal.psu.edu/cab38/ColorBrewer/ColorBrewer_updates.html",
    "http://schacon.github.com/git",
    "http://scitools.github.com/cartopy",
    "http://www.wmo.int/pages/prog/www/DPFS/documents/485_Vol_I_en_colour.pdf",
    "https://software.ac.uk/how-cite-software",
    "http://www.esrl.noaa.gov/psd/data/gridded/conventions/cdc_netcdf_standard.shtml",
]

# list of sources to exclude from the build.
exclude_patterns = []

# -- sphinx-gallery config ----------------------------------------------------
# See https://sphinx-gallery.github.io/stable/configuration.html

sphinx_gallery_conf = {
    # path to your example scripts
    "examples_dirs": ["../gallery_code"],
    # path to where to save gallery generated output
    "gallery_dirs": ["generated/gallery"],
    # filename pattern for the files in the gallery
    "filename_pattern": "/plot_",
    # filename patternt to ignore in the gallery
    "ignore_pattern": r"__init__\.py",
}


# -----------------------------------------------------------------------------
# Remove matplotlib agg warnings from generated doc when using plt.show
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message="Matplotlib is currently using agg, which is a"
    " non-GUI backend, so cannot show the figure.",
)


# -- numfig options (built-in) ------------------------------------------------
# Enable numfig.
numfig = True

numfig_format = {
    "code-block": "Example %s",
    "figure": "Figure %s",
    "section": "Section %s",
    "table": "Table %s",
}



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function
:func:`iris.analysis.cartography.gridcell_angles`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from cf_units import Unit
from iris.cube import Cube
from iris.coords import AuxCoord

from iris.analysis.cartography import gridcell_angles
from iris.tests.stock import sample_2d_latlons, lat_lon_cube


def _2d_multicells_testcube(cellsize_degrees=1.0):
    """
    Create a test cube with a grid of X and Y points, where each gridcell
    is independent (disjoint), arranged at an angle == the x-coord point.

    """
    # Setup np.linspace arguments to make the coordinate points.
    x0, x1, nx = -164, 164, 9
    y0, y1, ny = -75, 75, 7

    lats = np.linspace(y0, y1, ny, endpoint=True)
    lons_angles = np.linspace(x0, x1, nx, endpoint=True)
    x_pts_2d, y_pts_2d = np.meshgrid(lons_angles, lats)

    # Make gridcells rectangles surrounding these centrepoints, but also
    # tilted at various angles (= same as x-point lons, as that's easy).

    # Calculate centrepoint lons+lats : in radians, and shape (ny, nx, 1).
    xangs, yangs = np.deg2rad(x_pts_2d), np.deg2rad(y_pts_2d)
    xangs, yangs = [arr[..., None] for arr in (xangs, yangs)]
    # Program which corners are up+down on each gridcell axis.
    dx_corners = [[[-1, 1, 1, -1]]]
    dy_corners = [[[-1, -1, 1, 1]]]
    # Calculate the relative offsets in x+y at the 4 corners.
    x_ofs_2d = cellsize_degrees * np.cos(xangs) * dx_corners
    x_ofs_2d -= cellsize_degrees * np.sin(xangs) * dy_corners
    y_ofs_2d = cellsize_degrees * np.cos(xangs) * dy_corners
    y_ofs_2d += cellsize_degrees * np.sin(xangs) * dx_corners
    # Apply a latitude stretch to make correct angles on the globe.
    y_ofs_2d *= np.cos(yangs)
    # Make bounds arrays by adding the corner offsets to the centrepoints.
    x_bds_2d = x_pts_2d[..., None] + x_ofs_2d
    y_bds_2d = y_pts_2d[..., None] + y_ofs_2d

    # Create a cube with these points + bounds in its 'X' and 'Y' coords.
    co_x = AuxCoord(
        points=x_pts_2d,
        bounds=x_bds_2d,
        standard_name="longitude",
        units="degrees",
    )
    co_y = AuxCoord(
        points=y_pts_2d,
        bounds=y_bds_2d,
        standard_name="latitude",
        units="degrees",
    )
    cube = Cube(np.zeros((ny, nx)))
    cube.add_aux_coord(co_x, (0, 1))
    cube.add_aux_coord(co_y, (0, 1))
    return cube


class TestGridcellAngles(tests.IrisTest):
    def setUp(self):
        # Make a small "normal" contiguous-bounded cube to test on.
        # This one is regional.
        self.standard_regional_cube = sample_2d_latlons(
            regional=True, transformed=True
        )
        # Record the standard correct angle answers.
        result_cube = gridcell_angles(self.standard_regional_cube)
        result_cube.convert_units("degrees")
        self.standard_result_cube = result_cube
        self.standard_small_cube_results = result_cube.data

    def _check_multiple_orientations_and_latitudes(
        self,
        method="mid-lhs, mid-rhs",
        atol_degrees=0.005,
        cellsize_degrees=1.0,
    ):

        cube = _2d_multicells_testcube(cellsize_degrees=cellsize_degrees)

        # Calculate gridcell angles at each point.
        angles_cube = gridcell_angles(cube, cell_angle_boundpoints=method)

        # Check that the results are a close match to the original intended
        # gridcell orientation angles.
        # NOTE: neither the above gridcell construction nor the calculation
        # itself are exact :  Errors scale as the square of gridcell sizes.
        angles_cube.convert_units("degrees")
        angles_calculated = angles_cube.data

        # Note: the gridcell angles **should** just match the longitudes at
        # each point
        angles_expected = cube.coord("longitude").points

        # Wrap both into standard range for comparison.
        angles_calculated = (angles_calculated + 360.0) % 360.0
        angles_expected = (angles_expected + 360.0) % 360.0

        # Assert (toleranced) equality, and return results.
        self.assertArrayAllClose(
            angles_calculated, angles_expected, atol=atol_degrees
        )

        return angles_calculated, angles_expected

    def test_various_orientations_and_locations(self):
        self._check_multiple_orientations_and_latitudes()

    def test_result_form(self):
        # Check properties of the result cube *other than* the data values.
        test_cube = self.standard_regional_cube
        result_cube = self.standard_result_cube
        self.assertEqual(
            result_cube.long_name, "gridcell_angle_from_true_east"
        )
        self.assertEqual(result_cube.units, Unit("degrees"))
        self.assertEqual(len(result_cube.coords()), 2)
        self.assertEqual(
            result_cube.coord(axis="x"), test_cube.coord(axis="x")
        )
        self.assertEqual(
            result_cube.coord(axis="y"), test_cube.coord(axis="y")
        )

    def test_bottom_edge_method(self):
        # Get results with the "other" calculation method + check to tolerance.
        # A smallish cellsize should yield similar results in both cases.
        r1, _ = self._check_multiple_orientations_and_latitudes()
        r2, _ = self._check_multiple_orientations_and_latitudes(
            method="lower-left, lower-right",
            cellsize_degrees=0.1,
            atol_degrees=0.1,
        )

        # Not *exactly* the same : this checks we tested the 'other' method !
        self.assertFalse(np.allclose(r1, r2))
        # Note: results are a bit different in places.  This is acceptable.
        self.assertArrayAllClose(r1, r2, atol=0.1)

    def test_bounded_coord_args(self):
        # Check that passing the coords gives the same result as the cube.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        result = gridcell_angles(co_x, co_y)
        self.assertArrayAllClose(result.data, self.standard_small_cube_results)

    def test_coords_radians_args(self):
        # Check it still works with coords converted to radians.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        for coord in (co_x, co_y):
            coord.convert_units("radians")
        result = gridcell_angles(co_x, co_y)
        self.assertArrayAllClose(result.data, self.standard_small_cube_results)

    def test_bounds_array_args(self):
        # Check we can calculate from bounds values alone.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        # Results drawn from coord bounds should be nearly the same,
        # but not exactly, because of the different 'midpoint' values.
        result = gridcell_angles(co_x.bounds, co_y.bounds)
        self.assertArrayAllClose(
            result.data, self.standard_small_cube_results, atol=0.1
        )

    def test_unbounded_regional_coord_args(self):
        # Remove the coord bounds to check points-based calculation.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        for coord in (co_x, co_y):
            coord.bounds = None
        result = gridcell_angles(co_x, co_y)
        # Note: in this case, we can expect the leftmost and rightmost columns
        # to be rubbish, because the data is not global.
        # But the rest should match okay.
        self.assertArrayAllClose(
            result.data[:, 1:-1], self.standard_small_cube_results[:, 1:-1]
        )

    def test_points_array_args(self):
        # Check we can calculate from points arrays alone (no coords).
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        # As previous, the leftmost and rightmost columns are not good.
        result = gridcell_angles(co_x.points, co_y.points)
        self.assertArrayAllClose(
            result.data[:, 1:-1], self.standard_small_cube_results[:, 1:-1]
        )

    def test_unbounded_global(self):
        # For a contiguous global grid, a result based on points, i.e. with the
        # bounds removed, should be a reasonable match for the 'ideal' one
        # based on the bounds.

        # Make a global cube + calculate ideal bounds-based results.
        global_cube = sample_2d_latlons(transformed=True)
        result_cube = gridcell_angles(global_cube)
        result_cube.convert_units("degrees")
        global_cube_results = result_cube.data

        # Check a points-based calculation on the same basic grid.
        co_x, co_y = (global_cube.coord(axis=ax) for ax in ("x", "y"))
        for coord in (co_x, co_y):
            coord.bounds = None
        result = gridcell_angles(co_x, co_y)
        # In this case, the match is actually rather poor (!).
        self.assertArrayAllClose(result.data, global_cube_results, atol=7.5)
        # Leaving off first + last columns again gives a decent result.
        self.assertArrayAllClose(
            result.data[:, 1:-1], global_cube_results[:, 1:-1]
        )

        # NOTE: although this looks just as bad as 'test_points_array_args',
        # maximum errors there in the end columns are actually > 100 degrees !

    def test_nonlatlon_coord_system(self):
        # Check with points specified in an unexpected coord system.
        cube = sample_2d_latlons(regional=True, rotated=True)
        result = gridcell_angles(cube)
        self.assertArrayAllClose(result.data, self.standard_small_cube_results)
        # Check that the result has transformed (true-latlon) coordinates.
        self.assertEqual(len(result.coords()), 2)
        x_coord = result.coord(axis="x")
        y_coord = result.coord(axis="y")
        self.assertEqual(x_coord.shape, cube.shape)
        self.assertEqual(y_coord.shape, cube.shape)
        self.assertIsNotNone(cube.coord_system)
        self.assertIsNone(x_coord.coord_system)
        self.assertIsNone(y_coord.coord_system)

    def test_fail_coords_bad_units(self):
        # Check error with bad coords units.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        co_y.units = "m"
        with self.assertRaisesRegex(ValueError, "must have angular units"):
            gridcell_angles(co_x, co_y)

    def test_fail_nonarraylike(self):
        # Check error with bad args.
        co_x, co_y = 1, 2
        with self.assertRaisesRegex(
            ValueError, "must have array shape property"
        ):
            gridcell_angles(co_x, co_y)

    def test_fail_non2d_coords(self):
        # Check error with bad args.
        cube = lat_lon_cube()
        with self.assertRaisesRegex(
            ValueError, "inputs must have 2-dimensional shape"
        ):
            gridcell_angles(cube)

    def test_fail_different_shapes(self):
        # Check error with mismatched shapes.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        co_y = co_y[1:]
        with self.assertRaisesRegex(ValueError, "must have same shape"):
            gridcell_angles(co_x, co_y)

    def test_fail_different_coord_system(self):
        # Check error with mismatched coord systems.
        cube = sample_2d_latlons(regional=True, rotated=True)
        cube.coord(axis="x").coord_system = None
        with self.assertRaisesRegex(
            ValueError, "must have same coordinate system"
        ):
            gridcell_angles(cube)

    def test_fail_cube_dims(self):
        # Check error with mismatched cube dims.
        cube = self.standard_regional_cube
        # Make 5x6 into 5x5.
        cube = cube[:, :-1]
        co_x = cube.coord(axis="x")
        pts, bds = co_x.points, co_x.bounds
        co_new_x = co_x.copy(
            points=pts.transpose((1, 0)), bounds=bds.transpose((1, 0, 2))
        )
        cube.remove_coord(co_x)
        cube.add_aux_coord(co_new_x, (1, 0))
        with self.assertRaisesRegex(
            ValueError, "must have the same cube dimensions"
        ):
            gridcell_angles(cube)

    def test_fail_coord_noncoord(self):
        # Check that passing a coord + an array gives an error.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        with self.assertRaisesRegex(
            ValueError, "is a Coordinate, but .* is not"
        ):
            gridcell_angles(co_x, co_y.bounds)

    def test_fail_noncoord_coord(self):
        # Check that passing an array + a coord gives an error.
        co_x, co_y = (
            self.standard_regional_cube.coord(axis=ax) for ax in ("x", "y")
        )
        with self.assertRaisesRegex(
            ValueError, "is a Coordinate, but .* is not"
        ):
            gridcell_angles(co_x.points, co_y)

    def test_fail_bad_method(self):
        with self.assertRaisesRegex(
            ValueError, "unrecognised cell_angle_boundpoints"
        ):
            self._check_multiple_orientations_and_latitudes(
                method="something_unknown"
            )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Experimental cube-adjusting functions to assist merge operations.

"""


def equalise_attributes(cubes):
    """
    Delete cube attributes that are not identical over all cubes in a group.

    .. warning::

        This function is now **disabled**.

        The functionality has been moved to
        :func:`iris.util.equalise_attributes`.

    """
    old = "iris.experimental.equalise_cubes.equalise_attributes"
    new = "iris.util.equalise_attributes"
    emsg = (
        f'The function "{old}" has been moved.\n'
        f'Please replace "{old}(<cubes>)" with "{new}(<cubes>)".'
    )
    raise Exception(emsg)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides ABF (and ABL) file format capabilities.

ABF and ABL files are satellite file formats defined by Boston University.
Including this module adds ABF and ABL loading to the session's capabilities.

The documentation for this file format can be found
`here <http://cliveg.bu.edu/modismisr/lai3g-fpar3g.html>`_.

"""

import calendar
import datetime
import glob
import os.path

import numpy as np
import numpy.ma as ma

import iris
from iris.coords import AuxCoord, DimCoord
from iris.coord_systems import GeogCS
import iris.fileformats
import iris.io.format_picker


X_SIZE = 4320
Y_SIZE = 2160


month_numbers = {
    "jan": 1,
    "feb": 2,
    "mar": 3,
    "apr": 4,
    "may": 5,
    "jun": 6,
    "jul": 7,
    "aug": 8,
    "sep": 9,
    "oct": 10,
    "nov": 11,
    "dec": 12,
}


class ABFField:
    """
    A data field from an ABF (or ABL) file.

    Capable of creating a :class:`~iris.cube.Cube`.

    """

    def __init__(self, filename):
        """
        Create an ABFField object from the given filename.

        Args:

            * filename - An ABF filename.

        Example::

            field = ABFField("AVHRRBUVI01.1985feba.abl")

        """
        basename = os.path.basename(filename)
        if len(basename) != 24:
            raise ValueError(
                "ABFField expects a filename of 24 characters: "
                "{}".format(basename)
            )
        self._filename = filename

    def __getattr__(self, key):
        # Do we need to load now?
        if key == "data" and "data" not in self.__dict__:
            self._read()
        try:
            return self.__dict__[key]
        except KeyError:
            raise AttributeError("ABFField has no attribute '{}'".format(key))

    def _read(self):
        """Read the field from the given filename."""
        basename = os.path.basename(self._filename)
        self.version = int(basename[9:11])
        self.year = int(basename[12:16])
        self.month = basename[16:19]
        self.period = basename[19:20]
        self.format = basename[21:24]

        self.month = month_numbers[self.month]

        # Data is 8 bit bigendian.
        data = np.fromfile(self._filename, dtype=">u1").reshape(X_SIZE, Y_SIZE)
        # Iris' preferred dimensional ordering is (y,x).
        data = data.transpose()
        # Flip, for a positive step through the Y dimension.
        data = data[::-1]
        # Any percentages greater than 100 represent missing data.
        data = ma.masked_greater(data, 100)
        # The default fill value is 999999(!), so we choose something
        # more sensible. NB. 999999 % 256 = 63 = bad.
        data.fill_value = 255
        self.data = data

    def to_cube(self):
        """Return a new :class:`~iris.cube.Cube` from this ABFField."""

        cube = iris.cube.Cube(self.data)

        # Name.
        if self.format.lower() == "abf":
            cube.rename("FAPAR")
        elif self.format.lower() == "abl":
            cube.rename("leaf_area_index")
        else:
            msg = "Unknown ABF/ABL format: {}".format(self.format)
            raise iris.exceptions.TranslationError(msg)
        cube.units = "%"

        # Grid.
        step = 1.0 / 12.0

        llcs = GeogCS(semi_major_axis=6378137.0, semi_minor_axis=6356752.31424)

        x_coord = DimCoord(
            np.arange(X_SIZE) * step + (step / 2) - 180,
            standard_name="longitude",
            units="degrees",
            coord_system=llcs,
        )

        y_coord = DimCoord(
            np.arange(Y_SIZE) * step + (step / 2) - 90,
            standard_name="latitude",
            units="degrees",
            coord_system=llcs,
        )

        x_coord.guess_bounds()
        y_coord.guess_bounds()

        cube.add_dim_coord(x_coord, 1)
        cube.add_dim_coord(y_coord, 0)

        # Time.
        if self.period == "a":
            start = 1
            end = 15
        elif self.period == "b":
            start = 16
            end = calendar.monthrange(self.year, self.month)[1]
        else:
            raise iris.exceptions.TranslationError(
                "Unknown period: " "{}".format(self.period)
            )

        start = datetime.date(year=self.year, month=self.month, day=start)
        end = datetime.date(year=self.year, month=self.month, day=end)

        # Convert to "days since 0001-01-01".
        # Iris will have proper datetime objects in the future.
        # This step will not be necessary.
        start = start.toordinal() - 1
        end = end.toordinal() - 1

        # TODO: Should we put the point in the middle of the period instead?
        cube.add_aux_coord(
            AuxCoord(
                start,
                standard_name="time",
                units="days since 0001-01-01",
                bounds=[start, end],
            )
        )

        # TODO: Do they only come from Boston?
        # Attributes.
        cube.attributes["source"] = "Boston University"

        return cube


def load_cubes(filespecs, callback=None):
    """
    Loads cubes from a list of ABF filenames.

    Args:

    * filenames - list of ABF filenames to load

    Kwargs:

    * callback - a function that can be passed to :func:`iris.io.run_callback`

    .. note::

        The resultant cubes may not be in the same order as in the file.

    """
    if isinstance(filespecs, str):
        filespecs = [filespecs]

    for filespec in filespecs:
        for filename in glob.glob(filespec):

            field = ABFField(filename)
            cube = field.to_cube()

            # Were we given a callback?
            if callback is not None:
                cube = iris.io.run_callback(callback, cube, field, filename)
                if cube is None:
                    continue

            yield cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Statistical operations between cubes.

"""

import numpy as np
import numpy.ma as ma
import iris
from iris.util import broadcast_to_shape


def pearsonr(
    cube_a,
    cube_b,
    corr_coords=None,
    weights=None,
    mdtol=1.0,
    common_mask=False,
):
    """
    Calculate the Pearson's r correlation coefficient over specified
    dimensions.

    Args:

    * cube_a, cube_b (cubes):
        Cubes between which the correlation will be calculated.  The cubes
        should either be the same shape and have the same dimension coordinates
        or one cube should be broadcastable to the other.
    * corr_coords (str or list of str):
        The cube coordinate name(s) over which to calculate correlations. If no
        names are provided then correlation will be calculated over all common
        cube dimensions.
    * weights (numpy.ndarray, optional):
        Weights array of same shape as (the smaller of) cube_a and cube_b. Note
        that latitude/longitude area weights can be calculated using
        :func:`iris.analysis.cartography.area_weights`.
    * mdtol (float, optional):
        Tolerance of missing data. The missing data fraction is calculated
        based on the number of grid cells masked in both cube_a and cube_b. If
        this fraction exceed mdtol, the returned value in the corresponding
        cell is masked. mdtol=0 means no missing data is tolerated while
        mdtol=1 means the resulting element will be masked if and only if all
        contributing elements are masked in cube_a or cube_b. Defaults to 1.
    * common_mask (bool):
        If True, applies a common mask to cube_a and cube_b so only cells which
        are unmasked in both cubes contribute to the calculation. If False, the
        variance for each cube is calculated from all available cells. Defaults
        to False.

    Returns:
        A cube of the correlation between the two input cubes along the
        specified dimensions, at each point in the remaining dimensions of the
        cubes.

        For example providing two time/altitude/latitude/longitude cubes and
        corr_coords of 'latitude' and 'longitude' will result in a
        time/altitude cube describing the latitude/longitude (i.e. pattern)
        correlation at each time/altitude point.

    Reference:
        https://en.wikipedia.org/wiki/Pearson_correlation_coefficient

    This operation is non-lazy.

    """

    # Assign larger cube to cube_1
    if cube_b.ndim > cube_a.ndim:
        cube_1 = cube_b
        cube_2 = cube_a
    else:
        cube_1 = cube_a
        cube_2 = cube_b

    smaller_shape = cube_2.shape

    dim_coords_1 = [coord.name() for coord in cube_1.dim_coords]
    dim_coords_2 = [coord.name() for coord in cube_2.dim_coords]
    common_dim_coords = list(set(dim_coords_1) & set(dim_coords_2))
    # If no coords passed then set to all common dimcoords of cubes.
    if corr_coords is None:
        corr_coords = common_dim_coords

    def _ones_like(cube):
        # Return a copy of cube with the same mask, but all data values set to 1.
        # The operation is non-lazy.
        # For safety we also discard any cell-measures and ancillary-variables, to
        # avoid cube arithmetic possibly objecting to them, or inadvertently retaining
        # them in the result where they might be inappropriate.
        ones_cube = cube.copy()
        ones_cube.data = np.ones_like(cube.data)
        ones_cube.rename("unknown")
        ones_cube.units = 1
        for cm in ones_cube.cell_measures():
            ones_cube.remove_cell_measure(cm)
        for av in ones_cube.ancillary_variables():
            ones_cube.remove_ancillary_variable(av)
        return ones_cube

    # Match up data masks if required.
    if common_mask:
        # Create a cube of 1's with a common mask.
        if ma.is_masked(cube_2.data):
            mask_cube = _ones_like(cube_2)
        else:
            mask_cube = 1.0
        if ma.is_masked(cube_1.data):
            # Take a slice to avoid unnecessary broadcasting of cube_2.
            slice_coords = [
                dim_coords_1[i]
                for i in range(cube_1.ndim)
                if dim_coords_1[i] not in common_dim_coords
                and np.array_equal(
                    cube_1.data.mask.any(axis=i), cube_1.data.mask.all(axis=i)
                )
            ]
            cube_1_slice = next(cube_1.slices_over(slice_coords))
            mask_cube = _ones_like(cube_1_slice) * mask_cube
        # Apply common mask to data.
        if isinstance(mask_cube, iris.cube.Cube):
            cube_1 = cube_1 * mask_cube
            cube_2 = mask_cube * cube_2
            dim_coords_2 = [coord.name() for coord in cube_2.dim_coords]

    # Broadcast weights to shape of cubes if necessary.
    if weights is None or cube_1.shape == smaller_shape:
        weights_1 = weights
        weights_2 = weights
    else:
        if weights.shape != smaller_shape:
            raise ValueError(
                "weights array should have dimensions {}".format(smaller_shape)
            )

        dims_1_common = [
            i
            for i in range(cube_1.ndim)
            if dim_coords_1[i] in common_dim_coords
        ]
        weights_1 = broadcast_to_shape(weights, cube_1.shape, dims_1_common)
        if cube_2.shape != smaller_shape:
            dims_2_common = [
                i
                for i in range(cube_2.ndim)
                if dim_coords_2[i] in common_dim_coords
            ]
            weights_2 = broadcast_to_shape(
                weights, cube_2.shape, dims_2_common
            )
        else:
            weights_2 = weights

    # Calculate correlations.
    s1 = cube_1 - cube_1.collapsed(
        corr_coords, iris.analysis.MEAN, weights=weights_1
    )
    s2 = cube_2 - cube_2.collapsed(
        corr_coords, iris.analysis.MEAN, weights=weights_2
    )

    covar = (s1 * s2).collapsed(
        corr_coords, iris.analysis.SUM, weights=weights_1, mdtol=mdtol
    )
    var_1 = (s1 ** 2).collapsed(
        corr_coords, iris.analysis.SUM, weights=weights_1
    )
    var_2 = (s2 ** 2).collapsed(
        corr_coords, iris.analysis.SUM, weights=weights_2
    )

    denom = iris.analysis.maths.apply_ufunc(
        np.sqrt, var_1 * var_2, new_unit=covar.units
    )
    corr_cube = covar / denom
    corr_cube.rename("Pearson's r")

    return corr_cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the animation of cubes within iris.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

import iris
from iris.coord_systems import GeogCS

# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import iris.experimental.animate as animate
    import iris.plot as iplt


@tests.skip_plot
class IntegrationTest(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        cube = iris.cube.Cube(np.arange(36, dtype=np.int32).reshape((3, 3, 4)))
        cs = GeogCS(6371229)

        coord = iris.coords.DimCoord(
            points=np.array([1, 2, 3], dtype=np.int32), long_name="time"
        )
        cube.add_dim_coord(coord, 0)

        coord = iris.coords.DimCoord(
            points=np.array([-1, 0, 1], dtype=np.int32),
            standard_name="latitude",
            units="degrees",
            coord_system=cs,
        )
        cube.add_dim_coord(coord, 1)
        coord = iris.coords.DimCoord(
            points=np.array([-1, 0, 1, 2], dtype=np.int32),
            standard_name="longitude",
            units="degrees",
            coord_system=cs,
        )
        cube.add_dim_coord(coord, 2)
        self.cube = cube

    def test_cube_animation(self):
        # This follows :meth:`~matplotlib.animation.FuncAnimation.save`
        # to ensure that each frame corresponds to known accepted frames for
        # the animation.
        cube_iter = self.cube.slices(("latitude", "longitude"))

        ani = animate.animate(cube_iter, iplt.contourf)

        # Disconnect the first draw callback to stop the animation
        ani._fig.canvas.mpl_disconnect(ani._first_draw_id)

        ani = [ani]
        # Extract frame data
        for data in zip(*[a.new_saved_frame_seq() for a in ani]):
            # Draw each frame
            for anim, d in zip(ani, data):
                anim._draw_next_frame(d, blit=False)
                self.check_graphic()


if __name__ == "__main__":
    tests.main()



import matplotlib.pyplot as plt
import iris
import iris.plot as iplt


fname = iris.sample_data_path("air_temp.pp")

# Load exactly one cube from the given file
temperature = iris.load_cube(fname)

# We are only interested in a small number of longitudes (the 4 after and
# including the 5th element), so index them out
temperature = temperature[5:9, :]

for cube in temperature.slices("longitude"):

    # Create a string label to identify this cube (i.e. latitude: value)
    cube_label = "latitude: %s" % cube.coord("latitude").points[0]

    # Plot the cube, and associate it with a label
    iplt.plot(cube, label=cube_label)

# Match the longitude range to global
max_lon = temperature.coord("longitude").points.max()
min_lon = temperature.coord("longitude").points.min()
plt.xlim(min_lon, max_lon)

# Add the legend with 2 columns
plt.legend(ncol=2)

# Put a grid on the plot
plt.grid(True)

# Provide some axis labels
plt.ylabel("Temerature / kelvin")
plt.xlabel("Longitude / degrees")

# And a sensible title
plt.suptitle("Air Temperature", fontsize=20, y=0.9)

# Finally, show it.
plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp.load` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import iris.fileformats.pp as pp


class Test_load(tests.IrisTest):
    def test_call_structure(self):
        # Check that the load function calls the two necessary utility
        # functions.
        extract_result = mock.Mock()
        interpret_patch = mock.patch(
            "iris.fileformats.pp._interpret_fields",
            autospec=True,
            return_value=iter([]),
        )
        field_gen_patch = mock.patch(
            "iris.fileformats.pp._field_gen",
            autospec=True,
            return_value=extract_result,
        )
        with interpret_patch as interpret, field_gen_patch as field_gen:
            pp.load("mock", read_data=True)

        interpret.assert_called_once_with(extract_result)
        field_gen.assert_called_once_with(
            "mock", read_data_bytes=True, little_ended=False
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._convert_time_coords`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from cf_units import Unit, CALENDAR_GREGORIAN
from cftime import datetime as nc_datetime
import numpy as np
import unittest

from iris.coords import DimCoord, AuxCoord
from iris.fileformats.pp import SplittableInt
from iris.fileformats.pp_load_rules import _convert_time_coords
from iris.tests.unit.fileformats import TestField


def _lbtim(ia=0, ib=0, ic=0):
    return SplittableInt(ic + 10 * (ib + 10 * ia), {"ia": 2, "ib": 1, "ic": 0})


def _lbcode(value=None, ix=None, iy=None):
    if value is not None:
        result = SplittableInt(value, {"iy": slice(0, 2), "ix": slice(2, 4)})
    else:
        # N.B. if 'value' is None, both ix and iy must be set.
        result = SplittableInt(
            10000 + 100 * ix + iy, {"iy": slice(0, 2), "ix": slice(2, 4)}
        )
    return result


_EPOCH_HOURS_UNIT = Unit("hours since epoch", calendar=CALENDAR_GREGORIAN)
_HOURS_UNIT = Unit("hours")


class TestLBTIMx0x_SingleTimepoint(TestField):
    def _check_timepoint(self, lbcode, expect_match=True):
        lbtim = _lbtim(ib=0, ic=1)
        t1 = nc_datetime(1970, 1, 1, hour=6, minute=0, second=0)
        t2 = nc_datetime(0, 0, 0)  # not used in result
        lbft = None  # unused
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        if expect_match:
            expect_result = [
                (
                    DimCoord(
                        24 * 0.25,
                        standard_name="time",
                        units=_EPOCH_HOURS_UNIT,
                    ),
                    None,
                )
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)

    def test_normal_xy_dims(self):
        self._check_timepoint(_lbcode(1))

    def test_non_time_cross_section(self):
        self._check_timepoint(_lbcode(ix=1, iy=2))

    def test_time_cross_section(self):
        self._check_timepoint(_lbcode(ix=1, iy=20), expect_match=False)


class TestLBTIMx1x_Forecast(TestField):
    def _check_forecast(self, lbcode, expect_match=True):
        lbtim = _lbtim(ib=1, ic=1)
        # Validity time
        t1 = nc_datetime(1970, 1, 10, hour=6, minute=0, second=0)
        # Forecast time
        t2 = nc_datetime(1970, 1, 9, hour=3, minute=0, second=0)
        lbft = None  # unused
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        if expect_match:
            expect_result = [
                (
                    DimCoord(
                        24 * 1.125,
                        standard_name="forecast_period",
                        units="hours",
                    ),
                    None,
                ),
                (
                    DimCoord(
                        24 * 9.25,
                        standard_name="time",
                        units=_EPOCH_HOURS_UNIT,
                    ),
                    None,
                ),
                (
                    DimCoord(
                        24 * 8.125,
                        standard_name="forecast_reference_time",
                        units=_EPOCH_HOURS_UNIT,
                    ),
                    None,
                ),
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)

    def test_normal_xy(self):
        self._check_forecast(_lbcode(1))

    def test_non_time_cross_section(self):
        self._check_forecast(_lbcode(ix=1, iy=2))

    def test_time_cross_section(self):
        self._check_forecast(_lbcode(ix=1, iy=20), expect_match=False)

    def test_exact_hours(self):
        lbtim = _lbtim(ib=1, ic=1)
        t1 = nc_datetime(2015, 1, 20, hour=7, minute=0, second=0)
        t2 = nc_datetime(2015, 1, 20, hour=0, minute=0, second=0)
        coords_and_dims = _convert_time_coords(
            lbcode=_lbcode(1),
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=None,
        )
        (fp, _), (t, _), (frt, _) = coords_and_dims
        # These should both be exact whole numbers.
        self.assertEqual(fp.points[0], 7)
        self.assertEqual(t.points[0], 394927)

    def test_not_exact_hours(self):
        lbtim = _lbtim(ib=1, ic=1)
        t1 = nc_datetime(2015, 1, 20, hour=7, minute=10, second=0)
        t2 = nc_datetime(2015, 1, 20, hour=0, minute=0, second=0)
        coords_and_dims = _convert_time_coords(
            lbcode=_lbcode(1),
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=None,
        )
        (fp, _), (t, _), (frt, _) = coords_and_dims
        self.assertArrayAllClose(fp.points[0], 7.1666666, atol=0.0001, rtol=0)
        self.assertArrayAllClose(t.points[0], 394927.166666, atol=0.01, rtol=0)


class TestLBTIMx2x_TimePeriod(TestField):
    def _check_period(self, lbcode, expect_match=True):
        lbtim = _lbtim(ib=2, ic=1)
        # Start time
        t1 = nc_datetime(1970, 1, 9, hour=3, minute=0, second=0)
        # End time
        t2 = nc_datetime(1970, 1, 10, hour=3, minute=0, second=0)
        lbft = 2.0  # sample period
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        if expect_match:
            expect_result = [
                (
                    DimCoord(
                        24 * 9.125 - 2.0,
                        standard_name="forecast_reference_time",
                        units=_EPOCH_HOURS_UNIT,
                    ),
                    None,
                ),
                (
                    DimCoord(
                        standard_name="forecast_period",
                        units="hours",
                        points=[-10.0],
                        bounds=[-22.0, 2.0],
                    ),
                    None,
                ),
                (
                    DimCoord(
                        standard_name="time",
                        units=_EPOCH_HOURS_UNIT,
                        points=[24 * 8.625],
                        bounds=[24 * 8.125, 24 * 9.125],
                    ),
                    None,
                ),
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)

    def test_normal_xy(self):
        self._check_period(_lbcode(1))

    def test_non_time_cross_section(self):
        self._check_period(_lbcode(ix=1, iy=2))

    def test_time_cross_section(self):
        self._check_period(_lbcode(ix=1, iy=20), expect_match=False)


class TestLBTIMx3x_YearlyAggregation(TestField):
    def _check_yearly(self, lbcode, expect_match=True):
        lbtim = _lbtim(ib=3, ic=1)
        # Start time
        t1 = nc_datetime(1970, 1, 9, hour=9, minute=0, second=0)
        # End time
        t2 = nc_datetime(1972, 1, 11, hour=9, minute=0, second=0)
        lbft = 3.0  # sample period
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        if expect_match:
            t1_hours = 24 * 8.375
            t2_hours = 24 * (10.375 + 2 * 365)
            period_hours = 24.0 * (2 * 365 + 2)
            expect_result = [
                (
                    DimCoord(
                        [t2_hours - lbft],
                        standard_name="forecast_reference_time",
                        units=_EPOCH_HOURS_UNIT,
                    ),
                    None,
                ),
                (
                    DimCoord(
                        standard_name="forecast_period",
                        units="hours",
                        points=[lbft],
                        bounds=[lbft - period_hours, lbft],
                    ),
                    None,
                ),
                (
                    DimCoord(
                        standard_name="time",
                        units=_EPOCH_HOURS_UNIT,
                        points=[t2_hours],
                        bounds=[t1_hours, t2_hours],
                    ),
                    None,
                ),
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)

    def test_normal_xy(self):
        self._check_yearly(_lbcode(1))

    def test_non_time_cross_section(self):
        self._check_yearly(_lbcode(ix=1, iy=2))

    def test_time_cross_section(self):
        self._check_yearly(_lbcode(ix=1, iy=20), expect_match=False)


class TestLBTIMx2x_ZeroYear(TestField):
    def test_(self):
        lbtim = _lbtim(ib=2, ic=1)
        t1 = nc_datetime(0, 1, 1)
        t2 = nc_datetime(0, 1, 31, 23, 59, 00)
        lbft = 0
        lbcode = _lbcode(1)
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        self.assertEqual(coords_and_dims, [])


class TestLBTIMxxx_Unhandled(TestField):
    def test_unrecognised(self):
        lbtim = _lbtim(ib=4, ic=1)
        t1 = nc_datetime(0, 0, 0)
        t2 = nc_datetime(0, 0, 0)
        lbft = None
        lbcode = _lbcode(0)
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        self.assertEqual(coords_and_dims, [])


class TestLBCODE3xx(TestField):
    def test(self):
        lbcode = _lbcode(value=31323)
        lbtim = _lbtim(ib=2, ic=2)
        t1 = nc_datetime(1970, 1, 3, hour=0, minute=0, second=0)
        t2 = nc_datetime(1970, 1, 4, hour=0, minute=0, second=0)
        lbft = 24 * 4
        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
        )
        t2_hours = 24 * 3
        expected_result = [
            (
                DimCoord(
                    [t2_hours - lbft],
                    standard_name="forecast_reference_time",
                    units=_EPOCH_HOURS_UNIT,
                ),
                None,
            )
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected_result)


class TestArrayInputWithLBTIM_0_0_1(TestField):
    def test_t1_list(self):
        # lbtim ia = 0, ib = 0, ic = 1
        # with a series of times (t1).
        lbcode = _lbcode(1)
        lbtim = _lbtim(ia=0, ib=0, ic=1)
        hours = np.array([0, 3, 6, 9, 12])
        # Validity time - vector of different values
        t1 = [nc_datetime(1970, 1, 9, hour=3 + hour) for hour in hours]
        t1_dims = (0,)
        # Forecast reference time - scalar (not used)
        t2 = nc_datetime(1970, 1, 9, hour=3)
        lbft = None

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
        )

        # Expected coords.
        time_coord = DimCoord(
            (24 * 8) + 3 + hours, standard_name="time", units=_EPOCH_HOURS_UNIT
        )
        expected = [(time_coord, (0,))]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)


class TestArrayInputWithLBTIM_0_1_1(TestField):
    def test_t1_list_t2_scalar(self):
        # lbtim ia = 0, ib = 1, ic = 1
        # with a single forecast reference time (t2) and a series
        # of validity times (t1).
        lbcode = _lbcode(1)
        lbtim = _lbtim(ia=0, ib=1, ic=1)
        forecast_period_in_hours = np.array([0, 3, 6, 9, 12])
        # Validity time - vector of different values
        t1 = [
            nc_datetime(1970, 1, 9, hour=(3 + fp))
            for fp in forecast_period_in_hours
        ]
        t1_dims = (0,)
        # Forecast reference time - scalar
        t2 = nc_datetime(1970, 1, 9, hour=3)
        lbft = None  # Not used.

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
        )

        # Expected coords.
        fp_coord = DimCoord(
            forecast_period_in_hours,
            standard_name="forecast_period",
            units="hours",
        )
        time_coord = DimCoord(
            (24 * 8) + 3 + forecast_period_in_hours,
            standard_name="time",
            units=_EPOCH_HOURS_UNIT,
        )
        fref_time_coord = DimCoord(
            (24 * 8) + 3,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0,)),
            (time_coord, (0,)),
            (fref_time_coord, None),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)

    def test_t1_and_t2_list(self):
        # lbtim ia = 0, ib = 1, ic = 1
        # with a single repeated forecast reference time (t2) and a series
        # of validity times (t1).
        lbcode = _lbcode(1)
        lbtim = _lbtim(ia=0, ib=1, ic=1)
        forecast_period_in_hours = np.array([0, 3, 6, 9, 12])
        # Validity time - vector of different values
        t1 = [
            nc_datetime(1970, 1, 9, hour=(3 + fp))
            for fp in forecast_period_in_hours
        ]
        t1_dims = (0,)
        # Forecast reference time - vector of same values
        t2 = [
            nc_datetime(1970, 1, 9, hour=3) for _ in forecast_period_in_hours
        ]
        t2_dims = (0,)
        lbft = None  # Not used.

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
            t2_dims=t2_dims,
        )

        # Expected coords.
        fp_coord = DimCoord(
            forecast_period_in_hours,
            standard_name="forecast_period",
            units="hours",
        )
        time_coord = DimCoord(
            (24 * 8) + 3 + forecast_period_in_hours,
            standard_name="time",
            units=_EPOCH_HOURS_UNIT,
        )
        fref_time_coord = DimCoord(
            (24 * 8) + 3,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0,)),
            (time_coord, (0,)),
            (fref_time_coord, None),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)

    def test_t1_and_t2_orthogonal_lists(self):
        # lbtim ia = 0, ib = 1, ic = 1
        # with a single repeated forecast reference time (t2) and a series
        # of validity times (t1).
        lbcode = _lbcode(1)
        lbtim = _lbtim(ia=0, ib=1, ic=1)
        years = np.array([1970, 1971, 1972])
        hours = np.array([3, 6, 9, 12])
        # Validity time - vector of different values
        t1 = [nc_datetime(year, 1, 9, hour=12) for year in years]
        t1_dims = (0,)
        # Forecast reference time - vector of different values
        t2 = [nc_datetime(1970, 1, 9, hour=hour) for hour in hours]
        t2_dims = (1,)
        lbft = None  # Not used.

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
            t2_dims=t2_dims,
        )

        # Expected coords.
        points = [
            [(year - 1970) * 365 * 24 + 12 - hour for hour in hours]
            for year in years
        ]
        fp_coord = AuxCoord(
            points, standard_name="forecast_period", units="hours"
        )
        points = (years - 1970) * 24 * 365 + (24 * 8) + 12
        time_coord = DimCoord(
            points, standard_name="time", units=_EPOCH_HOURS_UNIT
        )
        points = (24 * 8) + hours
        fref_time_coord = DimCoord(
            points,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0, 1)),  # Spans dims 0 and 1.
            (time_coord, (0,)),
            (fref_time_coord, (1,)),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)

    def test_t1_multi_dim_list_t2_scalar(self):
        # Another case of lbtim ia = 0, ib = 1, ic = 1 but
        # with a changing forecast reference time (t2) and
        # validity time (t1).
        lbcode = _lbcode(1)
        lbtim = _lbtim(ia=0, ib=1, ic=1)
        forecast_period_in_hours = np.array([0, 3, 6, 9, 12])
        years = np.array([1970, 1971, 1972])
        # Validity time - 2d array of different values
        t1 = [
            [
                nc_datetime(year, 1, 9, hour=(3 + fp))
                for fp in forecast_period_in_hours
            ]
            for year in years
        ]
        t1_dims = (0, 1)
        # Forecast reference time - vector of different values
        t2 = nc_datetime(1970, 1, 9, hour=3)
        lbft = None  # Not used.

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
        )

        # Expected coords.
        fp_coord = AuxCoord(
            [
                forecast_period_in_hours + (year - 1970) * 365 * 24
                for year in years
            ],
            standard_name="forecast_period",
            units="hours",
        )
        time_coord = AuxCoord(
            [
                (24 * 8)
                + 3
                + forecast_period_in_hours
                + (year - 1970) * 365 * 24
                for year in years
            ],
            standard_name="time",
            units=_EPOCH_HOURS_UNIT,
        )
        fref_time_coord = DimCoord(
            (24 * 8) + 3,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0, 1)),
            (time_coord, (0, 1)),
            (fref_time_coord, None),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)

    def test_t1_and_t2_nparrays(self):
        # lbtim ia = 0, ib = 1, ic = 1
        # with a single repeated forecast reference time (t2) and a series
        # of validity times (t1).
        lbcode = _lbcode(1)
        lbtim = _lbtim(ia=0, ib=1, ic=1)
        forecast_period_in_hours = np.array([0, 3, 6, 9, 12])
        # Validity time - vector of different values
        t1 = np.array(
            [
                nc_datetime(1970, 1, 9, hour=(3 + fp))
                for fp in forecast_period_in_hours
            ]
        )
        t1_dims = (0,)
        # Forecast reference time - vector of same values
        t2 = np.array(
            [nc_datetime(1970, 1, 9, hour=3) for _ in forecast_period_in_hours]
        )
        t2_dims = (0,)
        lbft = None  # Not used.

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
            t2_dims=t2_dims,
        )

        # Expected coords.
        fp_coord = DimCoord(
            forecast_period_in_hours,
            standard_name="forecast_period",
            units="hours",
        )
        time_coord = DimCoord(
            (24 * 8) + 3 + forecast_period_in_hours,
            standard_name="time",
            units=_EPOCH_HOURS_UNIT,
        )
        fref_time_coord = DimCoord(
            (24 * 8) + 3,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0,)),
            (time_coord, (0,)),
            (fref_time_coord, None),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)


class TestArrayInputWithLBTIM_0_2_1(TestField):
    def test_t1_list_t2_scalar(self):
        lbtim = _lbtim(ib=2, ic=1)
        lbcode = _lbcode(1)
        hours = np.array([0, 3, 6, 9])
        # Start times - vector
        t1 = [nc_datetime(1970, 1, 9, hour=9 + hour) for hour in hours]
        t1_dims = (0,)
        # End time - scalar
        t2 = nc_datetime(1970, 1, 11, hour=9)
        lbft = 3.0  # Sample period

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t1_dims=t1_dims,
        )

        # Expected coords.
        points = lbft - (48 - hours) / 2.0
        bounds = np.array(
            [lbft - (48 - hours), np.ones_like(hours) * lbft]
        ).transpose()
        fp_coord = AuxCoord(
            points,
            standard_name="forecast_period",
            units="hours",
            bounds=bounds,
        )
        points = 9 * 24 + 9 + (hours / 2.0)
        bounds = np.array(
            [8 * 24 + 9 + hours, np.ones_like(hours) * 10 * 24 + 9]
        ).transpose()
        time_coord = AuxCoord(
            points,
            standard_name="time",
            units=_EPOCH_HOURS_UNIT,
            bounds=bounds,
        )
        points = 10 * 24 + 9 - lbft
        fref_time_coord = DimCoord(
            points,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0,)),
            (time_coord, (0,)),
            (fref_time_coord, None),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)


class TestArrayInputWithLBTIM_0_3_1(TestField):
    @unittest.skip("#3508 investigate unit test failure")
    def test_t1_scalar_t2_list(self):
        lbtim = _lbtim(ib=3, ic=1)
        lbcode = _lbcode(1)
        years = np.array([1972, 1973, 1974])
        # Start times - scalar
        t1 = nc_datetime(1970, 1, 9, hour=9)
        # End time - vector
        t2 = [nc_datetime(year, 1, 11, hour=9) for year in years]
        t2_dims = (0,)
        lbft = 3.0  # Sample period

        coords_and_dims = _convert_time_coords(
            lbcode=lbcode,
            lbtim=lbtim,
            epoch_hours_unit=_EPOCH_HOURS_UNIT,
            t1=t1,
            t2=t2,
            lbft=lbft,
            t2_dims=t2_dims,
        )

        # Expected coords.
        points = np.ones_like(years) * lbft
        bounds = np.array(
            [lbft - ((years - 1970) * 365 * 24 + 2 * 24), points]
        ).transpose()
        fp_coord = AuxCoord(
            points,
            standard_name="forecast_period",
            units="hours",
            bounds=bounds,
        )
        points = (years - 1970) * 365 * 24 + 10 * 24 + 9
        bounds = np.array(
            [np.ones_like(points) * (8 * 24 + 9), points]
        ).transpose()
        # The time coordinate is an AuxCoord as the lower bound for each
        # cell is the same so it does not meet the monotonicity requirement.
        time_coord = AuxCoord(
            points,
            standard_name="time",
            units=_EPOCH_HOURS_UNIT,
            bounds=bounds,
        )
        fref_time_coord = DimCoord(
            points - lbft,
            standard_name="forecast_reference_time",
            units=_EPOCH_HOURS_UNIT,
        )
        expected = [
            (fp_coord, (0,)),
            (time_coord, (0,)),
            (fref_time_coord, (0,)),
        ]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp._create_field_data` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

import iris.fileformats.pp as pp


class Test__create_field_data(tests.IrisTest):
    def test_loaded_bytes(self):
        # Check that a field with LoadedArrayBytes in core_data gets the
        # result of a suitable call to _data_bytes_to_shaped_array().
        mock_loaded_bytes = mock.Mock(spec=pp.LoadedArrayBytes)
        core_data = mock.MagicMock(return_value=mock_loaded_bytes)
        field = mock.Mock(core_data=core_data)
        data_shape = mock.Mock()
        land_mask = mock.Mock()
        with mock.patch(
            "iris.fileformats.pp._data_bytes_to_shaped_array"
        ) as convert_bytes:
            convert_bytes.return_value = mock.sentinel.array
            pp._create_field_data(field, data_shape, land_mask)

        self.assertIs(field.data, mock.sentinel.array)
        convert_bytes.assert_called_once_with(
            mock_loaded_bytes.bytes,
            field.lbpack,
            field.boundary_packing,
            data_shape,
            mock_loaded_bytes.dtype,
            field.bmdi,
            land_mask,
        )

    def test_deferred_bytes(self):
        # Check that a field with deferred array bytes in core_data gets a
        # dask array.
        fname = mock.sentinel.fname
        position = mock.sentinel.position
        n_bytes = mock.sentinel.n_bytes
        newbyteorder = mock.Mock(return_value=mock.sentinel.dtype)
        dtype = mock.Mock(newbyteorder=newbyteorder)
        deferred_bytes = (fname, position, n_bytes, dtype)
        core_data = mock.MagicMock(return_value=deferred_bytes)
        field = mock.Mock(core_data=core_data)
        data_shape = (100, 120)
        proxy = mock.Mock(
            dtype=np.dtype("f4"),
            shape=data_shape,
            spec=pp.PPDataProxy,
            ndim=len(data_shape),
        )
        # We can't directly inspect the concrete data source underlying
        # the dask array, so instead we patch the proxy creation and check it's
        # being created and invoked correctly.
        with mock.patch("iris.fileformats.pp.PPDataProxy") as PPDataProxy:
            PPDataProxy.return_value = proxy
            pp._create_field_data(field, data_shape, land_mask_field=None)
        # The data should be assigned via field.data. As this is a mock object
        # we can check the attribute directly.
        self.assertEqual(field.data.shape, data_shape)
        self.assertEqual(field.data.dtype, np.dtype("f4"))
        # Is it making use of a correctly configured proxy?
        # NB. We know it's *using* the result of this call because
        # that's where the dtype came from above.
        PPDataProxy.assert_called_once_with(
            (data_shape),
            dtype,
            fname,
            position,
            n_bytes,
            field.raw_lbpack,
            field.boundary_packing,
            field.bmdi,
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:meth:`iris.analysis.trajectory._nearest_neighbour_indices_ndcoords`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.cube import Cube
from iris.coords import DimCoord, AuxCoord

from iris.analysis.trajectory import (
    _nearest_neighbour_indices_ndcoords as nn_ndinds,
)


class Test2d(tests.IrisTest):
    def test_nonlatlon_simple_2d(self):
        co_y = DimCoord([10.0, 20.0], long_name="y")
        co_x = DimCoord([1.0, 2.0, 3.0], long_name="x")
        cube = Cube(np.zeros((2, 3)))
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        sample_point = [("x", 2.8), ("y", 18.5)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(1, 2)])

    def test_nonlatlon_multiple_2d(self):
        co_y = DimCoord([10.0, 20.0], long_name="y")
        co_x = DimCoord([1.0, 2.0, 3.0], long_name="x")
        cube = Cube(np.zeros((2, 3)))
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        sample_points = [("x", [2.8, -350.0, 1.7]), ("y", [18.5, 8.7, 12.2])]
        result = nn_ndinds(cube, sample_points)
        self.assertEqual(result, [(1, 2), (0, 0), (0, 1)])

    def test_latlon_simple_2d(self):
        co_y = DimCoord(
            [10.0, 20.0], standard_name="latitude", units="degrees"
        )
        co_x = DimCoord(
            [1.0, 2.0, 3.0], standard_name="longitude", units="degrees"
        )
        cube = Cube(np.zeros((2, 3)))
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        sample_point = [("longitude", 2.8), ("latitude", 18.5)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(1, 2)])

    def test_latlon_multiple_2d(self):
        co_y = DimCoord(
            [10.0, 20.0], standard_name="latitude", units="degrees"
        )
        co_x = DimCoord(
            [1.0, 2.0, 3.0], standard_name="longitude", units="degrees"
        )
        cube = Cube(np.zeros((2, 3)))
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        sample_points = [
            ("longitude", [2.8, -350.0, 1.7]),
            ("latitude", [18.5, 8.7, 12.2]),
        ]
        result = nn_ndinds(cube, sample_points)
        # Note slight difference from non-latlon version.
        self.assertEqual(result, [(1, 2), (0, 2), (0, 1)])


class Test1d(tests.IrisTest):
    def test_nonlatlon_simple_1d(self):
        co_x = AuxCoord([1.0, 2.0, 3.0, 1.0, 2.0, 3.0], long_name="x")
        co_y = AuxCoord([10.0, 10.0, 10.0, 20.0, 20.0, 20.0], long_name="y")
        cube = Cube(np.zeros(6))
        cube.add_aux_coord(co_y, 0)
        cube.add_aux_coord(co_x, 0)
        sample_point = [("x", 2.8), ("y", 18.5)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(5,)])

    def test_latlon_simple_1d(self):
        cube = Cube([11.0, 12.0, 13.0, 21.0, 22.0, 23.0])
        co_x = AuxCoord(
            [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
            standard_name="longitude",
            units="degrees",
        )
        co_y = AuxCoord(
            [10.0, 10.0, 10.0, 20.0, 20.0, 20.0],
            standard_name="latitude",
            units="degrees",
        )
        cube.add_aux_coord(co_y, 0)
        cube.add_aux_coord(co_x, 0)
        sample_point = [("longitude", 2.8), ("latitude", 18.5)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(5,)])


class TestApiExtras(tests.IrisTest):
    # Check operation with alternative calling setups.
    def test_no_y_dim(self):
        # Operate in X only, returned slice should be [:, ix].
        co_x = DimCoord([1.0, 2.0, 3.0], long_name="x")
        co_y = DimCoord([10.0, 20.0], long_name="y")
        cube = Cube(np.zeros((2, 3)))
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        sample_point = [("x", 2.8)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(slice(None), 2)])

    def test_no_x_dim(self):
        # Operate in Y only, returned slice should be [iy, :].
        co_x = DimCoord([1.0, 2.0, 3.0], long_name="x")
        co_y = DimCoord([10.0, 20.0], long_name="y")
        cube = Cube(np.zeros((2, 3)))
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        sample_point = [("y", 18.5)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(1, slice(None))])

    def test_sample_dictionary(self):
        # Pass sample_point arg as a dictionary: this usage mode is deprecated.
        co_x = AuxCoord([1.0, 2.0, 3.0], long_name="x")
        co_y = AuxCoord([10.0, 20.0], long_name="y")
        cube = Cube(np.zeros((2, 3)))
        cube.add_aux_coord(co_y, 0)
        cube.add_aux_coord(co_x, 1)
        sample_point = {"x": 2.8, "y": 18.5}
        exp_emsg = r"must be a list of \(coordinate, value\) pairs"
        with self.assertRaisesRegex(TypeError, exp_emsg):
            nn_ndinds(cube, sample_point)


class TestLatlon(tests.IrisTest):
    # Check correct calculations on lat-lon points.
    def _testcube_latlon_1d(self, lats, lons):
        cube = Cube(np.zeros(len(lons)))
        co_x = AuxCoord(lons, standard_name="longitude", units="degrees")
        co_y = AuxCoord(lats, standard_name="latitude", units="degrees")
        cube.add_aux_coord(co_y, 0)
        cube.add_aux_coord(co_x, 0)
        return cube

    def _check_latlon_1d(self, lats, lons, sample_point, expect):
        cube = self._testcube_latlon_1d(lats, lons)
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(expect,)])

    def test_lat_scaling(self):
        # Check that (88, 25) is closer to (88, 0) than to (87, 25)
        self._check_latlon_1d(
            lats=[88, 87],
            lons=[0, 25],
            sample_point=[("latitude", 88), ("longitude", 25)],
            expect=0,
        )

    def test_alternate_latlon_names_okay(self):
        # Check that (88, 25) is **STILL** closer to (88, 0) than to (87, 25)
        # ... when coords have odd, but still recognisable, latlon names.
        cube = self._testcube_latlon_1d(lats=[88, 87], lons=[0, 25])
        cube.coord("latitude").rename("y_latitude_y")
        cube.coord("longitude").rename("x_longitude_x")
        sample_point = [("y_latitude_y", 88), ("x_longitude_x", 25)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(0,)])

    def test_alternate_nonlatlon_names_different(self):
        # Check that (88, 25) is **NOT** closer to (88, 0) than to (87, 25)
        # ... by plain XY euclidean-distance, if coords have non-latlon names.
        cube = self._testcube_latlon_1d(lats=[88, 87], lons=[0, 25])
        cube.coord("latitude").rename("y")
        cube.coord("longitude").rename("x")
        sample_point = [("y", 88), ("x", 25)]
        result = nn_ndinds(cube, sample_point)
        self.assertEqual(result, [(1,)])

    def test_lons_wrap_359_0(self):
        # Check that (0, 359) is closer to (0, 0) than to (0, 350)
        self._check_latlon_1d(
            lats=[0, 0],
            lons=[0, 350],
            sample_point=[("latitude", 0), ("longitude", 359)],
            expect=0,
        )

    def test_lons_wrap_359_neg1(self):
        # Check that (0, 359) is closer to (0, -1) than to (0, 350)
        self._check_latlon_1d(
            lats=[0, 0],
            lons=[350, -1],
            sample_point=[("latitude", 0), ("longitude", 359)],
            expect=1,
        )

    def test_lons_wrap_neg179_plus179(self):
        # Check that (0, -179) is closer to (0, 179) than to (0, -170)
        self._check_latlon_1d(
            lats=[0, 0],
            lons=[-170, 179],
            sample_point=[("latitude", 0), ("longitude", -179)],
            expect=1,
        )

    def test_lons_over_pole(self):
        # Check that (89, 0) is closer to (89, 180) than to (85, 0)
        self._check_latlon_1d(
            lats=[85, 89],
            lons=[0, 180],
            sample_point=[("latitude", 89), ("longitude", 0)],
            expect=1,
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
A module to provide convenient file format identification through a combination of filename extension
and file based *magic* numbers.


To manage a collection of FormatSpecifications for loading::

    import iris.io.format_picker as fp
    import matplotlib.pyplot as plt
    fagent = fp.FormatAgent()
    png_spec = fp.FormatSpecification('PNG image', fp.MagicNumber(8),
                                      0x89504E470D0A1A0A,
                                      handler=lambda filename: plt.imread(filename),
                                      priority=5
                                      )
    fagent.add_spec(png_spec)

To identify a specific format from a file::

    with open(png_filename, 'rb') as png_fh:
        handling_spec = fagent.get_spec(png_filename, png_fh)

In the example, handling_spec will now be the png_spec previously added to the agent.

Now that a specification has been found, if a handler has been given with the specification, then the file can be handled::

    handler = handling_spec.handler
    if handler is None:
       raise ValueError('File cannot be handled.')
    else:
       result = handler(filename)

The calling sequence of handler is dependent on the function given in the original specification and can be customised to your project's needs.


"""

from collections.abc import Callable
import functools
import os
import struct


import iris.io


class FormatAgent:
    """
    The FormatAgent class is the containing object which is responsible for identifying the format of a given file
    by interrogating its children FormatSpecification instances.

    Typically a FormatAgent will be created empty and then extended with the :meth:`FormatAgent.add_spec` method::

        agent = FormatAgent()
        agent.add_spec(NetCDF_specification)

    Less commonly, this can also be written::

        agent = FormatAgent(NetCDF_specification)

    """

    def __init__(self, format_specs=None):
        """ """
        self._format_specs = list(format_specs or [])
        self._format_specs.sort()

    def add_spec(self, format_spec):
        """Add a FormatSpecification instance to this agent for format consideration."""
        self._format_specs.append(format_spec)
        self._format_specs.sort()

    def __repr__(self):
        return "FormatAgent(%r)" % self._format_specs

    def __str__(self):
        prefix = " * " if len(self._format_specs) > 1 else ""
        return prefix + "\n * ".join(
            ["%s" % format_spec for format_spec in self._format_specs]
        )

    def get_spec(self, basename, buffer_obj):
        """
        Pick the first FormatSpecification which can handle the given
        filename and file/buffer object.

        .. note::

            ``buffer_obj`` may be ``None`` when a seekable file handle is not
            feasible (such as over the http protocol). In these cases only the
            format specifications which do not require a file handle are
            tested.

        """
        element_cache = {}
        for format_spec in self._format_specs:
            # For the case where a buffer_obj is None (such as for the
            # http protocol) skip any specs which require a fh - they
            # don't match.
            if buffer_obj is None and format_spec.file_element.requires_fh:
                continue

            fmt_elem = format_spec.file_element
            fmt_elem_value = format_spec.file_element_value

            # cache the results for each file element
            if repr(fmt_elem) not in element_cache:
                # N.B. File oriented as this is assuming seekable stream.
                if buffer_obj is not None and buffer_obj.tell() != 0:
                    # reset the buffer if tell != 0
                    buffer_obj.seek(0)

                element_cache[repr(fmt_elem)] = fmt_elem.get_element(
                    basename, buffer_obj
                )

            # If we have a callable object, then call it and tests its result, otherwise test using basic equality
            if isinstance(fmt_elem_value, Callable):
                matches = fmt_elem_value(element_cache[repr(fmt_elem)])
            elif element_cache[repr(fmt_elem)] == fmt_elem_value:
                matches = True
            else:
                matches = False

            if matches:
                return format_spec

        printable_values = {}
        for key, value in element_cache.items():
            value = str(value)
            if len(value) > 50:
                value = value[:50] + "..."
            printable_values[key] = value
        msg = (
            "No format specification could be found for the given buffer."
            " File element cache:\n {}".format(printable_values)
        )
        raise ValueError(msg)


@functools.total_ordering
class FormatSpecification:
    """
    Provides the base class for file type definition.

    Every FormatSpecification instance has a name which can be accessed with the :attr:`FormatSpecification.name` property and
    a FileElement, such as filename extension or 32-bit magic number, with an associated value for format identification.

    """

    def __init__(
        self,
        format_name,
        file_element,
        file_element_value,
        handler=None,
        priority=0,
        constraint_aware_handler=False,
    ):
        """
        Constructs a new FormatSpecification given the format_name and particular FileElements

        Args:

        * format_name - string name of fileformat being described
        * file_element - FileElement instance of the element which identifies this FormatSpecification
        * file_element_value - The value that the file_element should take if a file matches this FormatSpecification

        Kwargs:

        * handler - function which will be called when the specification has been identified and is required to handler a format.
                            If None, then the file can still be identified but no handling can be done.
        * priority - Integer giving a priority for considering this specification where higher priority means sooner consideration.

        """
        if not isinstance(file_element, FileElement):
            raise ValueError(
                "file_element must be an instance of FileElement, got %r"
                % file_element
            )

        self._file_element = file_element
        self._file_element_value = file_element_value
        self._format_name = format_name
        self._handler = handler
        self.priority = priority
        self.constraint_aware_handler = constraint_aware_handler

    def __hash__(self):
        # Hashed by specification for consistent ordering in FormatAgent (including self._handler in this hash
        # for example would order randomly according to object id)
        return hash(self._file_element)

    @property
    def file_element(self):
        return self._file_element

    @property
    def file_element_value(self):
        return self._file_element_value

    @property
    def name(self):
        """The name of this FileFormat. (Read only)"""
        return self._format_name

    @property
    def handler(self):
        """The handler function of this FileFormat. (Read only)"""
        return self._handler

    def _sort_key(self):
        return (-self.priority, self.name, self.file_element)

    def __lt__(self, other):
        if not isinstance(other, FormatSpecification):
            return NotImplemented

        return self._sort_key() < other._sort_key()

    def __eq__(self, other):
        if not isinstance(other, FormatSpecification):
            return NotImplemented

        return self._sort_key() == other._sort_key()

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        # N.B. loader is not always going to provide a nice repr if it is a lambda function, hence a prettier version is available in __str__
        return "FormatSpecification(%r, %r, %r, handler=%r, priority=%s)" % (
            self._format_name,
            self._file_element,
            self._file_element_value,
            self.handler,
            self.priority,
        )

    def __str__(self):
        return "%s%s (priority %s)" % (
            self.name,
            " (no handler available)" if self.handler is None else "",
            self.priority,
        )


class FileElement:
    """
    Represents a specific aspect of a FileFormat which can be identified using the given element getter function.

    """

    def __init__(self, requires_fh=True):
        """
        Constructs a new file element, which may require a file buffer.

        Kwargs:

        * requires_fh - Whether this FileElement needs a file buffer.

        """
        self.requires_fh = requires_fh

    def get_element(self, basename, file_handle):
        """Called when identifying the element of a file that this FileElement is representing."""
        raise NotImplementedError("get_element must be defined in a subclass")

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)


class MagicNumber(FileElement):
    """A :class:`FileElement` that returns a byte sequence in the file."""

    len_formats = {4: ">L", 8: ">Q"}

    def __init__(self, num_bytes, offset=None):
        FileElement.__init__(self)
        self._num_bytes = num_bytes
        self._offset = offset

    def get_element(self, basename, file_handle):
        if self._offset is not None:
            file_handle.seek(self._offset)
        bytes = file_handle.read(self._num_bytes)
        fmt = self.len_formats.get(self._num_bytes)
        if len(bytes) != self._num_bytes:
            raise EOFError(file_handle.name)
        if fmt is None:
            result = bytes
        else:
            result = struct.unpack(fmt, bytes)[0]
        return result

    def __repr__(self):
        return "MagicNumber({}, {})".format(self._num_bytes, self._offset)


class FileExtension(FileElement):
    """A :class:`FileElement` that returns the extension from the filename."""

    def get_element(self, basename, file_handle):
        return os.path.splitext(basename)[1]


class LeadingLine(FileElement):
    """A :class:`FileElement` that returns the first line from the file."""

    def get_element(self, basename, file_handle):
        return file_handle.readline()


class UriProtocol(FileElement):
    """
    A :class:`FileElement` that returns the "scheme" and "part" from a URI,
    using :func:`~iris.io.decode_uri`.

    """

    def __init__(self):
        FileElement.__init__(self, requires_fh=False)

    def get_element(self, basename, file_handle):
        return iris.io.decode_uri(basename)[0]



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris.fileformats.nimrod_load_rules.units` function.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock
import numpy as np

from iris.cube import Cube
from iris.fileformats.nimrod_load_rules import (
    units,
    NIMROD_DEFAULT,
)
from iris.fileformats.nimrod import NimrodField


class Test(tests.IrisTest):
    NIMROD_LOCATION = "iris.fileformats.nimrod_load_rules"

    def setUp(self):
        self.field = mock.Mock(
            units="",
            int_mdi=-32767,
            float32_mdi=NIMROD_DEFAULT,
            spec=NimrodField,
        )
        self.cube = Cube(np.ones((3, 3), dtype=np.float32))

    def _call_units(self, data=None, units_str=None):
        if data is not None:
            self.cube.data = data
        if units_str:
            self.field.units = units_str
        units(self.cube, self.field)

    def test_null(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(units_str="m")
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "m")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )

    def test_times32(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(
                data=np.ones_like(self.cube.data) * 32, units_str="mm/hr*32"
            )
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "mm/hr")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_visibility_units(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(
                data=((np.ones_like(self.cube.data) / 2) - 25000),
                units_str="m/2-25k",
            )
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "m")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_power_in_units(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(
                data=np.ones_like(self.cube.data) * 1000, units_str="mm*10^3"
            )
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "mm")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_ug_per_m3_units(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(
                data=((np.ones_like(self.cube.data) * 10)),
                units_str="ug/m3E1",
            )
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "ug/m3")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_g_per_kg(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(
                data=((np.ones_like(self.cube.data) * 1000)), units_str="g/Kg"
            )
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "kg/kg")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_unit_expection_dictionary(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(units_str="mb")
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "hPa")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_per_second(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(units_str="/s")
        self.assertEqual(warn.call_count, 0)
        self.assertEqual(self.cube.units, "s^-1")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)

    def test_unhandled_unit(self):
        with mock.patch("warnings.warn") as warn:
            self._call_units(units_str="kittens")
        self.assertEqual(warn.call_count, 1)
        self.assertEqual(self.cube.units, "")
        self.assertArrayAlmostEqual(
            self.cube.data, np.ones_like(self.cube.data)
        )
        self.assertEqual(self.cube.data.dtype, np.float32)
        self.assertEqual(self.cube.attributes["invalid_units"], "kittens")


if __name__ == "__main__":
    tests.main()
